import numpy as np
import pandas as pd

from . import accessor


@accessor.register_obscollection_accessor("geo")
class GeoAccessor:
    def __init__(self, oc_obj):
        self._obj = oc_obj

    def get_bounding_box(self, xcol="x", ycol="y", buffer=0):
        """returns the bounding box of all observations.

        Parameters
        ----------
        xcol : str, optional
            column name with x values
        ycol : str, optional
            column name with y values
        buffer : int or float, optional
            add a buffer around the bouding box from the observations

        Returns
        -------
        xmin, ymin, xmax, ymax
            coordinates of bouding box
        """

        xmin = self._obj[xcol].min() - buffer
        xmax = self._obj[xcol].max() + buffer
        ymin = self._obj[ycol].min() - buffer
        ymax = self._obj[ycol].max() + buffer

        return (xmin, ymin, xmax, ymax)

    def get_extent(self, xcol="x", ycol="y", buffer=0):
        """returns the extent of all observations.

        Parameters
        ----------
        xcol : str, optional
            column name with x values
        ycol : str, optional
            column name with y values
        buffer : int or float, optional
            add a buffer around the bouding box from the observations

        Returns
        -------
        xmin, xmax, ymin, ymax
            coordinates of bouding box
        """

        xmin = self._obj[xcol].min() - buffer
        xmax = self._obj[xcol].max() + buffer
        ymin = self._obj[ycol].min() - buffer
        ymax = self._obj[ycol].max() + buffer

        return (xmin, xmax, ymin, ymax)

    def set_lat_lon(self, in_epsg="epsg:28992", out_epsg="epsg:4326", add_to_meta=True):
        """create columns with lat and lon values of the observation points.

        Parameters
        ----------
        in_epsg : str, optional
            epsg code of current x and y attributes, default (RD new)
        out_epsg : str, optional
            epsg code of desired output, default lat/lon
        add_to_meta : bool, optional
            if True the lat and lon values are added to the observation meta
            dictionary. The default is True.

        Returns
        -------
        None.
        """

        df_lat_lon = self._obj.geo.get_lat_lon(in_epsg, out_epsg)
        for iname in df_lat_lon.index:
            self._obj._set_metadata_value(
                iname, "lat", df_lat_lon.loc[iname, "lat"], add_to_meta
            )
            self._obj._set_metadata_value(
                iname, "lon", df_lat_lon.loc[iname, "lon"], add_to_meta
            )

    def get_lat_lon(self, in_epsg="epsg:28992", out_epsg="epsg:4326"):
        """get lattitude and longitude from x and y attributes.

        Parameters
        ----------
        in_epsg : str, optional
            epsg code of current x and y attributes, default (RD new)
        out_epsg : str, optional
            epsg code of desired output, default lat/lon

        Returns
        -------
        pandas.DataFrame
            with columns 'lat' and 'lon'
        """

        df_lat_lon = pd.DataFrame(index=self._obj.index, columns=["lat", "lon"])
        for iname in self._obj.index:
            o = self._obj.loc[iname, "obs"]
            df_lat_lon.loc[iname, ["lat", "lon"]] = o.geo.get_lat_lon(in_epsg, out_epsg)

        return df_lat_lon

    def get_nearest_point(
        self,
        obs_collection2=None,
        gdf2=None,
        xcol_obs1="x",
        ycol_obs1="y",
        xcol_obs2="x",
        ycol_obs2="y",
    ):
        """get nearest point of another obs collection for each point in the
        current obs collection.

        Parameters
        ----------
        obs_collection2 : ObsCollection, optional
            collection of observations of which the nearest point is found
        gdf2 : GeoDataFrame, optional
            dataframe to look for nearest observation point
        xcol_obs1 : str, optional
            x column in self._obj used to get geometry
        ycol_obs1 : str, optional
            y column in self._obj used to get geometry
        xcol_obs2 : str, optional
            x column in obs_collection2 used to get geometry
        ycol_obs2 : str, optional
            y column in self._obj used to get geometry

        Returns
        -------
        pandas.DataFrame
            with columns 'nearest point' and 'distance nearest point'
        """

        from shapely.ops import nearest_points

        gdf1 = self._obj.to_gdf(xcol=xcol_obs1, ycol=ycol_obs1)
        if obs_collection2 is not None:
            gdf2 = obs_collection2.to_gdf(xcol=xcol_obs2, ycol=ycol_obs2)
        elif gdf2 is None:
            raise ValueError("obs_collecction2 or gdf2 should be defined")

        pts_gdf2 = gdf2.geometry.unary_union

        def nearest_point(point_gdf1, pts=pts_gdf2):
            # find the nearest point and return the corresponding Place value
            nearest_point_gdf2 = nearest_points(point_gdf1, pts_gdf2)[1]
            nearest = gdf2[gdf2.geometry == nearest_point_gdf2]
            return nearest.index[0]

        def distance_nearest_point(point_gdf1, pts=pts_gdf2):
            # find the nearest point and return the corresponding Place value
            nearest_point_gdf2 = nearest_points(point_gdf1, pts_gdf2)[1]
            distance = point_gdf1.distance(nearest_point_gdf2)
            return distance

        gdf1["nearest point"] = gdf1.apply(
            lambda row: nearest_point(row.geometry), axis=1
        )
        gdf1["distance nearest point"] = gdf1.apply(
            lambda row: distance_nearest_point(row.geometry), axis=1
        )

        return gdf1[["nearest point", "distance nearest point"]]

    def _get_nearest_geometry(
        self,
        gdf=None,
        xcol_obs="x",
        ycol_obs="y",
        geometry_type="polygon",
        multiple_geometries="error",
    ):
        """get nearest geometry for each point in the obs collection. Function
        works for line and polygon geometries.

        Parameters
        ----------
        gdf : GeoDataFrame
            dataframe with polygon features
        xcol_obs : str, optional
            x column in self._obj used to get geometry
        ycol_obs : str, optional
            y column in self._obj used to get geometry
        geometry_type : str, optional
            geometry type, can be 'polygon' or 'line'.
        multiple_geometries : str, optional
            keyword on how to deal with multiple geometries being nearest.
            Options are:
                'error' -> raise a ValueError
                'keep_all' -> return the indices of multiple geometries as a
                string seperated by a comma
                'keep_first' -> return the index of the first geometry

        Returns
        -------
        pandas.DataFrame
            with columns 'nearest geometry' and 'distance nearest geometry'
        """

        gdf_obs = self._obj.to_gdf(xcol=xcol_obs, ycol=ycol_obs)
        gdf = gdf.copy()
        for i, point in gdf_obs.geometry.items():
            distances = [point.distance(pol) for pol in gdf.geometry.values]
            if (np.array(distances) == np.min(distances)).sum() > 1:
                if multiple_geometries == "error":
                    raise ValueError(f"multiple {geometry_type}s are nearest")
                elif multiple_geometries == "keep_all":
                    ids = []
                    for i_min in np.where(np.array(distances) == np.min(distances))[0]:
                        ids.append(gdf.index[i_min])
                    gdf_obs.loc[i, f"nearest {geometry_type}"] = ", ".join(ids)
                    gdf_obs.loc[i, f"distance nearest {geometry_type}"] = np.min(
                        distances
                    )
                elif multiple_geometries == "keep_first":
                    gdf_obs.loc[i, f"nearest {geometry_type}"] = gdf.iloc[
                        np.argmin(distances)
                    ].name
                    gdf_obs.loc[i, f"distance nearest {geometry_type}"] = np.min(
                        distances
                    )
                else:
                    raise ValueError(
                        "invalid value for multiple_geometries ->"
                        f"{multiple_geometries}"
                    )
            else:
                gdf_obs.loc[i, f"nearest {geometry_type}"] = gdf.iloc[
                    np.argmin(distances)
                ].name
                gdf_obs.loc[i, f"distance nearest {geometry_type}"] = np.min(distances)

        return gdf_obs[
            [f"nearest {geometry_type}", f"distance nearest {geometry_type}"]
        ]

    def get_nearest_line(
        self, gdf=None, xcol_obs="x", ycol_obs="y", multiple_lines="error"
    ):
        """get nearest line for each point in the obs collection. Function
        calls the nearest_polygon function.

        Parameters
        ----------
        gdf : GeoDataFrame
            dataframe with line features
        xcol_obs : str, optional
            x column in self._obj used to get geometry
        ycol_obs : str, optional
            y column in self._obj used to get geometry
        multiple_lines : str, optional
            keyword on how to deal with multiple lines being nearest.
            Options are:
                'error' -> raise a ValueError
                'keep_all' -> return the indices of multiple lines as a
                string seperated by a comma
                'keep_first' -> return the index of the first line

        Returns
        -------
        pandas.DataFrame
            with columns 'nearest polygon' and 'distance nearest polygon'
        """
        return self._get_nearest_geometry(
            gdf=gdf,
            xcol_obs=xcol_obs,
            ycol_obs=ycol_obs,
            multiple_geometries=multiple_lines,
            geometry_type="line",
        )

    def get_nearest_polygon(
        self, gdf=None, xcol_obs="x", ycol_obs="y", multiple_polygons="error"
    ):
        """get nearest polygon for each point in the obs collection. Function
        also works for lines instead of polygons.

        Parameters
        ----------
        gdf : GeoDataFrame
            dataframe with polygon features
        xcol_obs : str, optional
            x column in self._obj used to get geometry
        ycol_obs : str, optional
            y column in self._obj used to get geometry
        multiple_polygons : str, optional
            keyword on how to deal with multiple polygons being nearest.
            Options are:
                'error' -> raise a ValueError
                'keep_all' -> return the indices of multiple polygons as a
                string seperated by a comma
                'keep_first' -> return the index of the first polygon

        Returns
        -------
        pandas.DataFrame
            with columns 'nearest polygon' and 'distance nearest polygon'
        """
        return self._get_nearest_geometry(
            gdf=gdf,
            xcol_obs=xcol_obs,
            ycol_obs=ycol_obs,
            multiple_geometries=multiple_polygons,
            geometry_type="polygon",
        )

    def get_distance_to_point(self, point, xcol="x", ycol="y"):
        """get distance of every observation to a point.

        Parameters
        ----------
        point : shapely.geometry.point.Point
            point geometry
        xcol : str, optional
            x column in self._obj used to get x coordinates
        ycol : str, optional
            y column in self._obj used to get y coordinates

        Returns
        -------
        pd.Series
            distance to the point for every observation in self._obj
        """

        gdf = self._obj[[xcol, ycol]].to_gdf(xcol=xcol, ycol=ycol)

        return gdf.distance(point)

    def within_extent(self, extent, inplace=False):
        """Slice ObsCollection by extent.

        Parameters
        ----------
        extent : tuple
            format (xmin, xmax, ymin, ymax), default dis.sr.get_extent() format

        Returns
        -------
        new_oc : obs_collection.ObsCollection
            ObsCollection with observations within extent
        """

        new_oc = self._obj[
            (self._obj.x > extent[0])
            & (self._obj.x < extent[1])
            & (self._obj.y > extent[2])
            & (self._obj.y < extent[3])
        ]
        if inplace:
            self._obj._update_inplace(new_oc)
        else:
            return new_oc

    def within_polygon(self, gdf=None, shapefile=None, inplace=False, **kwargs):
        """Slice ObsCollection by checking if points are within a shapefile.

        Parameters
        ----------
        gdf : GeoDataFrame, optional
            geodataframe containing a single polygon
        shapefile : str, optional
            Not yet implemented
        inplace : bool, default False
            Modify the ObsCollection in place (do not create a new object).
        **kwargs :
            kwargs will be passed to the self._obj.to_gdf() method

        Returns
        -------
        new_oc : obs_collection.ObsCollection
            ObsCollection with observations within polygon
        """

        if gdf is not None:
            if gdf.shape[0] != 1:
                raise NotImplementedError("cannot handle zero or multiple polygons")
            gdf_oc = self._obj.to_gdf(**kwargs)
            new_oc = self._obj.loc[gdf_oc.within(gdf.geometry.values[0])]
        elif shapefile is not None:
            raise NotImplementedError("shapefiles are not yet implemented")
        else:
            raise ValueError("shapefile or gdf must be specified")

        if inplace:
            self._obj._update_inplace(new_oc)
            return
        else:
            return new_oc


@accessor.register_obs_accessor("geo")
class GeoAccessorObs:
    def __init__(self, obs):
        self._obj = obs

    def get_lat_lon(self, in_epsg="epsg:28992", out_epsg="epsg:4326"):
        """get lattitude and longitude from x and y attributes.

        Parameters
        ----------
        in_epsg : str, optional
            epsg code of current x and y attributes, default (RD new)
        out_epsg : str, optional
            epsg code of desired output, default lat/lon

        Returns
        -------
        lon, lat : longitude and lattitude of x, y coordinates
        """

        from pyproj import Transformer

        transformer = Transformer.from_crs(in_epsg, out_epsg)

        if np.isnan(self._obj.x) or np.isnan(self._obj.y):
            return np.nan, np.nan

        return transformer.transform(self._obj.x, self._obj.y)



import test_001_to_from as ttf


def test_within_extent():
    dino_gw = ttf.test_obscollection_dinozip_gw()
    extent = [210350, 213300, 473300, 474000]
    dino_gw.geo.within_extent(extent, inplace=True)
    assert dino_gw.shape[0] == 4
    return


# %% stats


def test_obscollection_consecutive_obs_years():
    gw = ttf.test_obscollection_dinozip_gw_keep_all_obs()
    gw.stats.consecutive_obs_years()
    return


def test_obscollection_get_number_of_obs():
    gw = ttf.test_obscollection_dinozip_gw_keep_all_obs()
    gw.stats.get_no_of_observations()
    return


def test_obscollection_get_first_last_obs_date():
    gw = ttf.test_obscollection_dinozip_gw_keep_all_obs()
    gw.stats.get_first_last_obs_date()
    return


def test_obscollection_get_seasonal_stats():
    gw = ttf.test_obscollection_dinozip_gw_keep_all_obs()
    gw.stats.get_seasonal_stat(stat="mean")
    return


def test_obscollection_get_min():
    gw = ttf.test_obscollection_dinozip_gw_keep_all_obs()
    gw.stats.get_min()
    return


def test_obscollection_get_max():
    gw = ttf.test_obscollection_dinozip_gw_keep_all_obs()
    gw.stats.get_max()
    return


# %% geo


def test_get_nearest_point():
    # check two of the same observation collections
    # every point must find itself as the nearest point
    dino_gw = ttf.test_obscollection_dinozip_gw()
    fl = ttf.test_obscollection_dinozip_wl()
    dino_gw[
        ["nearest point", "distance nearest point"]
    ] = dino_gw.geo.get_nearest_point(fl)
    return


def test_get_nearest_polygon():
    import geopandas as gpd
    from shapely.geometry import Polygon

    # check two of the same observation collections
    # every point must find itself as the nearest point
    dino_gw = ttf.test_obscollection_dinozip_gw()
    extent = dino_gw.geo.get_extent()
    polygon1 = Polygon(
        (
            (extent[0], extent[2]),
            (extent[0], extent[3]),
            (extent[1], extent[3]),
            (extent[1], extent[2]),
            (extent[0], extent[2]),
        )
    )
    polygon2 = Polygon(
        (
            (-extent[0], -extent[2]),
            (-extent[0], -extent[3]),
            (-extent[1], -extent[3]),
            (-extent[1], -extent[2]),
            (-extent[0], -extent[2]),
        )
    )
    gdf = gpd.GeoDataFrame({"name": [1, 2], "geometry": [polygon1, polygon2]})

    dino_gw[
        ["nearest polygon", "distance nearest polygon"]
    ] = dino_gw.geo.get_nearest_polygon(gdf)
    assert (dino_gw["nearest polygon"] == 0.0).all()
    assert (dino_gw["distance nearest polygon"] == 0.0).all()

    return


def test_get_ground_level_oc():
    try:
        from art_tools import hpd_extension  # noqa: F401

        gw = ttf.test_obscollection_fews_lowmemory()
        gw.art.geo_get_ground_level()
        return
    except ModuleNotFoundError as e:
        print(e)
    return


def test_get_ground_level_gwobs():
    try:
        from art_tools import hpd_extension  # noqa: F401

        gw = ttf.test_observation_gw()
        gw.art.geo_get_ground_level()
        return
    except ModuleNotFoundError as e:
        print(e)
        return



import logging
import os

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.gridspec import GridSpec

from . import accessor

logger = logging.getLogger(__name__)


@accessor.register_obscollection_accessor("plots")
class CollectionPlots:
    def __init__(self, oc_obj):
        """Object containing plotting methods for ObsCollections.

        Parameters
        ----------
        oc : ObsCollection
            ObsCollection instance
        """
        self._obj = oc_obj

    def interactive_plots(
        self,
        savedir="figures",
        tmin=None,
        tmax=None,
        per_monitoring_well=True,
        **kwargs,
    ):
        """Create interactive plots of the observations using bokeh.

        Parameters
        ----------
        savedir : str
            directory used for the folium map and bokeh plots
        tmin : dt.datetime, optional
            start date for timeseries plot
        tmax : dt.datetime, optional
            end date for timeseries plot
        per_monitoring_well : bool, optional
            if True plot multiple tubes at the same monitoring_well in one
            figure
        **kwargs :
            will be passed to the Obs.to_interactive_plot method, options
            include:

            - cols : list of str or None
            - hoover_names : list of str
            - plot_freq : list of str
            - plot_legend_names : list of str
            - markers : list of str
            - hoover_names : list of str
            - plot_colors : list of str
            - ylabel : str
            - add_filter_to_legend : boolean
        """
        _color_cycle = (
            "blue",
            "olive",
            "lime",
            "red",
            "orange",
            "yellow",
            "purple",
            "silver",
            "powderblue",
            "salmon",
            "tan",
        )

        # check if observations consist of monitoring wells
        if per_monitoring_well:
            otype = self._obj._infer_otype()
            if isinstance(otype, (list, np.ndarray)):
                per_monitoring_well = False
            elif otype.__name__ == "GroundwaterObs":
                pass
            else:
                per_monitoring_well = False

        if per_monitoring_well:
            plot_names = self._obj.groupby("monitoring_well").count().index
        else:
            plot_names = self._obj.index

        for name in plot_names:
            if per_monitoring_well:
                oc = self._obj.loc[
                    self._obj.monitoring_well == name, "obs"
                ].sort_index()
            else:
                oc = self._obj.loc[[name], "obs"]

            p = None
            for i, o in enumerate(oc.values):
                if i == 10:
                    raise NotImplementedError(
                        "cannot add more than 10 lines to a single plot"
                    )
                try:
                    p = o.plots.interactive_plot(
                        savedir=savedir,
                        p=p,
                        tmin=tmin,
                        tmax=tmax,
                        plot_colors=[_color_cycle[i + 1]],
                        return_filename=False,
                        **kwargs,
                    )
                    logger.debug(f"created iplot -> {o.name}")
                except ValueError:
                    logger.error(f"{o.name} has no data between {tmin} and {tmax}")
                    o.iplot_fname = None

    def interactive_map(
        self,
        plot_dir="figures",
        m=None,
        tiles="OpenStreetMap",
        fname=None,
        per_monitoring_well=True,
        color="blue",
        legend_name=None,
        add_legend=True,
        map_label="",
        map_label_size=20,
        col_name_lat="lat",
        col_name_lon="lon",
        zoom_start=13,
        create_interactive_plots=True,
        **kwargs,
    ):
        """Create an interactive map with interactive plots using folium and
        bokeh.

        Notes
        -----
        Some notes on this method:

        - if you want to have multiple obs collections on one folium map, only
          the last one should have add_legend = True to create a correct legend
        - the color of the observation point on the map is now the same color
          as the line of the observation measurements. Also a built-in color
          cycle is used for different measurements at the same monitoring_well.

        Parameters
        ----------
        plot_dir : str
            directory used for the folium map and bokeh plots
        m : folium.Map, str, optional
            current map to add observations too, if None a new map is created
        tiles : str, optional
            background tiles, default is openstreetmap
        fname : str, optional
            name of the folium map
        per_monitoring_well : bool, optional
            if True plot multiple tubes at the same monitoring well in one
            figure
        color : str, optional
            color of the observation points on the map
        legend_name : str, optional
            the name of the observation points shown in the map legend
        add_legend : boolean, optional
            add a legend to a plot
        map_label : str, optional
            add a label to the monitoring wells on the map, the label should be
            1. the attribute of an observation
            2. the key in the meta attribute of the observation
            3. a generic label for each observation in this collection.
            A label is only added if map_label is not ''. The default is ''.
        map_label_size : int, optional
            label size of the map_label in pt.
        col_name_lat : str, optional
            name of the column in the obs_collection dic with the lat values
            of the observation points
        col_name_lon : str, optional
            see col_name_lat
        zoom_start : int, optional
            start zoom level of the folium ma
        create_interactive_plots : boolean, optional
            if True interactive plots will be created, if False the iplot_fname
            attribute of the observations is used.
        **kwargs :
            will be passed to the to_interactive_plots method options are:

            - cols : list of str or None
            - hoover_names : list of str
            - plot_legend_names : list of str
            - plot_freq : list of str
            - markers : list of str
            - hoover_names : list of str
            - plot_colors : list of str
            - ylabel : str
            - add_filter_to_legend : boolean
            - tmin : dt.datetime
            - tmax : dt.datetime

        Returns
        -------
        m : folium.Map
            the folium map
        """

        import branca
        import folium
        from folium.features import DivIcon

        # create interactive bokeh plots
        if create_interactive_plots:
            self._obj.plots.interactive_plots(
                savedir=plot_dir, per_monitoring_well=per_monitoring_well, **kwargs
            )

        # check if observation collection has lat and lon values
        if (col_name_lat not in self._obj.columns) and (
            col_name_lon not in self._obj.columns
        ):
            self._obj.geo.set_lat_lon()

        # determine start location of map
        northing = np.mean(
            (self._obj[col_name_lat].min(), self._obj[col_name_lat].max())
        )
        easting = np.mean(
            (self._obj[col_name_lon].min(), self._obj[col_name_lon].max())
        )

        # create map if no map is given
        if m is None:
            m = folium.Map([northing, easting], zoom_start=zoom_start)

        # get oc name if no legend name is given
        if legend_name is None:
            legend_name = self._obj.name

        # add the point observations with plots to the map
        group_name = '<span style=\\"color: {};\\">{}</span>'.format(color, legend_name)
        group = folium.FeatureGroup(name=group_name)

        # check if observations consist of monitoring wells
        if per_monitoring_well:
            otype = self._obj._infer_otype()
            if isinstance(otype, (list, np.ndarray)):
                per_monitoring_well = False
            elif otype.__name__ == "GroundwaterObs":
                pass
            else:
                per_monitoring_well = False

        if per_monitoring_well:
            plot_names = self._obj.groupby("monitoring_well").count().index
        else:
            plot_names = self._obj.index

        for name in plot_names:
            if per_monitoring_well:
                oc = self._obj.loc[
                    self._obj.monitoring_well == name, "obs"
                ].sort_index()
                o = oc.iloc[-1]
                name = o.name
            else:
                o = self._obj.loc[name, "obs"]

            if o.iplot_fname is not None:
                with open(o.iplot_fname, "r") as f:
                    bokeh_html = f.read()

                iframe = branca.element.IFrame(html=bokeh_html, width=620, height=420)
                popup = folium.Popup(iframe, max_width=620)

                folium.CircleMarker(
                    [
                        self._obj.loc[o.name, col_name_lat],
                        self._obj.loc[o.name, col_name_lon],
                    ],
                    icon=folium.Icon(icon="signal"),
                    fill=True,
                    color=color,
                    popup=popup,
                ).add_to(group)

                if map_label != "":
                    if map_label in o._metadata:
                        map_label_val = getattr(o, map_label)
                    elif map_label in o.meta.keys():
                        map_label_val = o.meta[map_label]
                    else:
                        map_label_val = map_label
                    folium.map.Marker(
                        [
                            self._obj.loc[name, col_name_lat],
                            self._obj.loc[name, col_name_lon],
                        ],
                        icon=DivIcon(
                            icon_size=(150, 36),
                            icon_anchor=(0, 0),
                            html='<div style="font-size: %ipt">%s</div>'
                            % (map_label_size, map_label_val),
                        ),
                    ).add_to(group)
            else:
                logger.info(f"no iplot available for {o.name}")

        group.add_to(m)

        # add legend
        if add_legend:
            folium.map.LayerControl("topright", collapsed=False).add_to(m)

        # save map
        # filename and path
        if fname is not None:
            if not fname.endswith(".html"):
                fname = fname + ".html"
            if not os.path.exists(plot_dir):
                os.mkdir(plot_dir)
            m.save(os.path.join(plot_dir, fname))

        return m

    def section_plot(
        self,
        tmin=None,
        tmax=None,
        cols=(None,),
        section_colname_x=None,
        section_label_x=None,
        section_well_layout_color="gray",
        section_markersize=100,
        ylabel="auto",
        fn_save=None,
        check_obs_close_to_screen_bottom=True,
        plot_well_layout_markers=True,
    ):
        """Create plot with well layout (left) en observations (right).

        Parameters
        ----------
        tmin : dt.datetime, optional
            start date for timeseries plot
        tmax : dt.datetime, optional
            end date for timeseries plot
        cols : tuple of str or None, optional
                the columns of the observation to plot. The first numeric column
                is used if cols is None, by default None.
        section_colname_x : str, optional
            column used for x position on section plot, when None order
            collection is used
        section_label_x : str, optional
            label applied to x-axis in section plot
        section_well_layout_color : str, optional
            color of well layout, default is gray
        section_markersize : int, optional
            size of makers in sectionplot
        ylabel: str or list, optional
            when 'auto' column unit in collection is ylabel, otherwise first
            element of list is label of section plot, second element of observation plot
        fn_save : str, optional
            filename to save plot
        check_obs_close_to_screen_bottom : bool, optional
            plots a horizontal line when minimum observation is close to screen_bottom
        plot_well_layout_markers : bool, optional
            plots ground level, top tube, screen levels and sandtrap via
            makers. Default is True

        TODO:
            - speficy colors via extra column in ObsCollection
            - addtional visual checks:
                maximum observation is close to or above ground level,
                maximum observation is close to or above tube top
                minimum observation is close to or below tube bottom (sand trap)
            - include some interactive Bokeh fancy
            - apply an offset when two wells are at same location
            - limit y-axis of section plot to observations only
            - remove the checking (if obs are near bottom) from this function
            - moving the legend outside the plot
            - set xlim of observation plot more tight when tmin is not specified
        """

        # prepare column for x location in section plot
        if section_colname_x is None:
            # use order in ObsCollection
            plot_x = np.arange(len(self._obj))
        else:
            plot_x = self._obj[section_colname_x]

        # create figure
        fig = plt.figure(figsize=(15, 5))
        gs = GridSpec(
            1,
            2,
            width_ratios=[1, 3],
        )
        ax_section = fig.add_subplot(gs[0])
        ax_obs = fig.add_subplot(gs[1])
        axes = [ax_section, ax_obs]

        if plot_well_layout_markers:
            # plot well layout via markers
            ax_section.scatter(
                plot_x,
                self._obj.tube_top.values,
                section_markersize,
                label="tube top",
                marker="*",
                facecolors="none",
                color=section_well_layout_color,
            )
            ax_section.scatter(
                plot_x,
                self._obj.ground_level.values,
                section_markersize,
                label="ground level",
                marker="_",
                color=section_well_layout_color,
            )
            ax_section.scatter(
                plot_x,
                self._obj.screen_top.values,
                section_markersize / 2,
                label="screen top",
                marker="x",
                color=section_well_layout_color,
            )
            ax_section.scatter(
                plot_x,
                self._obj.screen_bottom.values,
                section_markersize,
                label="screen bottom",
                marker="+",
                color=section_well_layout_color,
            )

        # loop over all wells, plot observations and details in section plot
        for counter, name in enumerate(self._obj.index):
            # which column to plot?
            cols = list(cols)
            for i, col in enumerate(cols):
                if col is None:
                    cols[i] = self._obj.loc[name, "obs"]._get_first_numeric_col_name()

            # create plot dataframe
            plot_df = self._obj.loc[name, "obs"][tmin:tmax][cols].copy()
            if plot_df.empty or plot_df[cols].isna().all().all():
                logger.warning(f"{name} has no data between {tmin} and {tmax}")
                continue

            # PART 1: plot timeseries of observations
            # one or multiple columns to plot?
            if len(cols) == 1:
                p = ax_obs.plot(plot_df, label=name)
            else:
                for col in cols:
                    p = ax_obs.plot(plot_df[col], label=f"{name}, {col}")

            if check_obs_close_to_screen_bottom:
                # add horizonal line to plot when minimum observation in first plot
                # column is close to bottom of screen
                offset = 0.1
                if self._obj.loc[name, "screen_bottom"] > (
                    plot_df[cols[0]].dropna().min() - offset
                ):
                    ax_obs.axhline(
                        y=self._obj.loc[name, "screen_bottom"],
                        ls="--",
                        lw=4,
                        alpha=0.5,
                        label=(
                            f"screen bottom of {name}\n"
                            "is close to minimum observation"
                        ),
                        color=p[0].get_color(),
                    )

            # PART 2: fancy section plot with lines along tube

            # highlight filter on section plot
            ax_section.plot(
                [plot_x[counter]] * 2,
                [
                    self._obj.loc[name, "screen_top"],
                    self._obj.loc[name, "screen_bottom"],
                ],
                color="k",
                lw=3,
                ls="-",
            )

            # highlight blind tube on section plot
            ax_section.plot(
                [plot_x[counter]] * 2,
                [self._obj.loc[name, "screen_top"], self._obj.loc[name, "tube_top"]],
                color=p[0].get_color(),
                lw=3,
            )

            # add sandtrap when present
            if "tube_bottom" in self._obj.columns:
                ax_section.plot(
                    [plot_x[counter]] * 2,
                    [
                        self._obj.loc[name, "screen_bottom"],
                        self._obj.loc[name, "tube_bottom"],
                    ],
                    color=p[0].get_color(),
                    lw=3,
                )

            # PART 3: fancy section plot with bandwith of observations
            ax_section.scatter(
                plot_x[counter],
                plot_df.quantile(q=0.95),
                section_markersize,
                marker=(3, 0, 0),
                color="blue",
                alpha=0.5,
                label="95% observation",
                zorder=100,
            )
            ax_section.scatter(
                plot_x[counter],
                plot_df.median(),
                section_markersize,
                marker=(4, 0, 90),
                color="green",
                alpha=0.5,
                label="median",
                zorder=100,
            )
            ax_section.scatter(
                plot_x[counter],
                plot_df.quantile(q=0.05),
                section_markersize,
                marker=(3, 0, 180),
                color="red",
                alpha=0.5,
                label="5% observation",
                zorder=100,
            )

        logger.info(f"created sectionplot -> {name}")

        # layout
        if section_label_x is None:
            # use name as xtick and rotate
            ax_section.set_xticks(
                plot_x,
                self._obj.index,
                rotation="vertical",
                fontsize="small",
            )
        else:
            ax_section.set_xlabel(section_label_x)

        ax_obs.set_xlim(left=tmin, right=tmax)

        # rotate labels on observation axis
        ax_obs.set_xticks(
            ax_obs.get_xticks(),
            ax_obs.get_xticklabels(),
            rotation="vertical",
            fontsize="small",
        )

        if ylabel == "auto":
            # has collection uniform unit?
            if len(self._obj.unit.unique()) == 1:
                ylabel = self._obj.unit.unique()[0]
            else:
                logger.error(
                    "Collection has more than one unit. Plot has both on one y-axis."
                )
                ylabel = ", ".join(map(str, self._obj.unit.unique()))
            for ax in axes:
                ax.set_ylabel(ylabel)
        else:
            try:
                ax_section.set_ylabel(ylabel[0])
                ax_obs.set_ylabel(ylabel[1])
            except ValueError:
                logger.error(f"Invalid value for ylabel {ylabel}. Plot has no ylabels.")

        # add layout to both plots
        for ax in axes:
            ax.grid()

            # simple legend
            handles, labels = ax.get_legend_handles_labels()
            by_label = dict(zip(labels, handles))
            ax.legend(by_label.values(), by_label.keys())

        fig.suptitle(self._obj.name)

        if fn_save is not None:
            try:
                fig.tight_layout()
                fig.savefig(
                    fn_save,
                    dpi=250,
                )
            except ValueError:
                logger.error(f"Save of figure {name} failed ({fn_save}).")

        return fig, axes


@accessor.register_obs_accessor("plots")
class ObsPlots:
    def __init__(self, obs):
        self._obj = obs

    def interactive_plot(
        self,
        savedir=None,
        cols=(None,),
        markers=("line",),
        p=None,
        plot_legend_names=("",),
        plot_freq=(None,),
        tmin=None,
        tmax=None,
        hoover_names=("Peil",),
        hoover_date_format="%Y-%m-%d",
        ylabel=None,
        plot_colors=("blue",),
        add_screen_to_legend=False,
        return_filename=False,
    ):
        """Create an interactive plot of the observations using bokeh.

        Todo:

        - add options for hoovers, markers, linestyle

        Parameters
        ----------
        savedir : str, optional
            directory used for the folium map and bokeh plots
        cols : tuple of str or None, optional
            the columns of the observation to plot. The first numeric column
            is used if cols is None, by default None.
        markers : list of str, optional
            type of markers that can be used for plot, 'line' and 'circle' are
            supported
        p : bokeh.plotting.figure, optional
            reference to existing figure, if p is None a new figure is created
        plot_legend_names : list of str, optional
            legend in bokeh plot
        plot_freq : list of str, optional
            bokeh plot is resampled with this frequency to reduce the size
        tmin : dt.datetime, optional
            start date for timeseries plot
        tmax : dt.datetime, optional
            end date for timeseries plot
        hoover_names : list of str, optional
            names will be displayed together with the cols values
            when hoovering over plot
        hoover_date_format : str, optional
            date format to use when hoovering over a plot
        ylabel : str or None, optional
            label on the y-axis. If None the unit attribute of the observation
            is used.
        plot_colors : list of str, optional
            plot_colors used for the plots
        add_screen_to_legend : boolean, optional
            if True the attributes screen_top and screen_bottom
            are added to the legend name
        return_filename : boolean, optional
            if True filename will be returned

        Returns
        -------
        fname_plot : str or bokeh plot
            filename of the bokeh plot or reference to bokeh plot
        """

        from bokeh.models import ColumnDataSource, HoverTool
        from bokeh.plotting import figure, save
        from bokeh.resources import CDN

        cols = list(cols)
        for i, col in enumerate(cols):
            if col is None:
                cols[i] = self._obj._get_first_numeric_col_name()

        # create plot dataframe
        plot_df = self._obj[tmin:tmax].copy()
        plot_df["date"] = plot_df.index.strftime(hoover_date_format)
        if plot_df.empty or plot_df[cols].isna().all().all():
            raise ValueError(
                "{} has no data between {} and {}".format(self._obj.name, tmin, tmax)
            )

        # create plot
        if p is None:
            p = figure(width=600, height=400, x_axis_type="datetime", title="")
            if ylabel is None:
                ylabel = self._obj.unit
            p.yaxis.axis_label = ylabel

        # get x axis
        xcol = self._obj.index.name
        if xcol is None:
            xcol = "index"

        # get color
        if len(plot_colors) < len(cols):
            plot_colors = list(plot_colors) * len(cols)

        # get base for hoover tooltips
        plots = []
        tooltips = []
        tooltips.append(("date", "@date"))

        # plot multiple columns
        for i, column in enumerate(cols):
            # legend name
            if add_screen_to_legend:
                lname = "{} {} (NAP {:.2f} - {:.2f})".format(
                    plot_legend_names[i],
                    self._obj.name,
                    self._obj.screen_bottom,
                    self._obj.screen_top,
                )
            else:
                lname = "{} {}".format(plot_legend_names[i], self._obj.name)

            # resample data
            if plot_freq[i] is None:
                source = ColumnDataSource(plot_df[[column, "date"]])
            else:
                source = ColumnDataSource(
                    plot_df[[column, "date"]].resample(plot_freq[i]).first()
                )

            # plot data

            if markers[i] in ["line", "l"]:
                plots.append(
                    p.line(
                        xcol,
                        column,
                        source=source,
                        color=plot_colors[i],
                        legend_label=lname,
                        alpha=0.8,
                        muted_alpha=0.2,
                    )
                )
            elif markers[i] in ["circle", "c"]:
                plots.append(
                    p.circle(
                        xcol,
                        column,
                        source=source,
                        color=plot_colors[i],
                        legend_label=lname,
                        alpha=0.8,
                        muted_alpha=0.2,
                    )
                )
            else:
                raise NotImplementedError(
                    "marker '{}' invalid. Only line and"
                    "circle are currently available".format(markers[i])
                )

            # add columns to hoover tooltips
            tooltips_p = tooltips.copy()
            tooltips_p.append((hoover_names[i], "@{}".format(column)))
            hover = HoverTool(renderers=[plots[i]], tooltips=tooltips_p, mode="vline")
            p.add_tools(hover)

        p.legend.location = "top_left"
        p.legend.click_policy = "mute"

        # save plot
        if savedir is not None:
            if not os.path.isdir(savedir):
                os.makedirs(savedir)
            self._obj.iplot_fname = os.path.join(savedir, self._obj.name + ".html")
            save(p, self._obj.iplot_fname, resources=CDN, title=self._obj.name)

        if return_filename:
            return self._obj.iplot_fname
        else:
            return p



import logging

import numpy as np
import pandas as pd

from . import accessor

logger = logging.getLogger(__name__)


def get_model_layer_z(z, zvec, left=-999, right=999):
    """Get index of model layer based on elevation.

    Assumptions:

    - the highest value in zvec is the top of model layer 0.
    - if z is equal to the bottom of a layer, the model layer above that
      layer is assigned.

    Parameters
    ----------
    z : int, float
        elevation.
    zvec : list, np.array
        elevations of model layers. shape is nlay + 1
    left : int, optional
        if z is below the lowest value in zvec, this value is returned.
        The default is -999.
    right : TYPE, optional
        if z is above the highest value in zvec, this value is returned.
        The default is 999.

    Returns
    -------
    int : int
        model layer

    Examples
    --------
    >>> zvec = [0, -10, -20, -30]
    >>> get_model_layer_z(-5, zvec)
    0

    >>> get_model_layer_z(-25, zvec)
    2

    >>> get_model_layer_z(-50, zvec)
    -999

    >>> get_model_layer_z(100, zvec)
    999

    >>> get_model_layer_z(-20, zvec)
    1
    """
    # make sure zvec is descending
    zvec = np.sort(zvec)[::-1]
    if z in zvec:
        z += 1e-10
    lay = int(
        np.interp(z, zvec[::-1], np.arange(len(zvec))[::-1], left=left, right=right)
    )

    return lay


def check_if_var_is_invalid(var):
    if var is None:
        return True
    elif np.isnan(var):
        return True

    return False


def get_modellayer_from_screen_depth(ftop, fbot, zvec, left=-999, right=999):
    """


    Parameters
    ----------
    ftop : int or float
        top of screen.
    fbot : int or float
        bottom of screen, has to be lower than ftop.
    zvec : list or np.array
        elevations of the modellayers at the location of the tube.
    left : int, optional
        value to return if tube screen is below the modellayers.
        The default is -999.
    right : int, optional
        value to return if tube screen is above the modellayers.
        The default is-999.

    Raises
    ------
    ValueError
        raised if something unexpected happens.

    Returns
    -------
    int or np.nan
        modellayer.

    examples
    --------
    >>> zvec = [0, -10, -20, -30, -40]
    >>> get_modellayer_from_screen_depth(-5, -7, zvec)
    0

    >>> get_modellayer_from_screen_depth(-25, -27, zvec)
    2

    >>> get_modellayer_from_screen_depth(-15, -27, zvec)
    2

    >>> get_modellayer_from_screen_depth(-5, -27, zvec)
    1

    >>> get_modellayer_from_screen_depth(-5, -37, zvec)
    1

    >>> get_modellayer_from_screen_depth(15, -5, zvec)
    0

    >>> get_modellayer_from_screen_depth(15, 5, zvec)
    999

    >>> get_modellayer_from_screen_depth(-55, -65, zvec)
    -999

    >>> get_modellayer_from_screen_depth(15, -65, zvec)
    nan

    >>> get_modellayer_from_screen_depth(None, -7, zvec)
    0

    >>> get_modellayer_from_screen_depth(None, None, zvec)
    nan

    """

    zvec = np.sort(zvec)[::-1]
    ftop_invalid = check_if_var_is_invalid(ftop)
    fbot_invalid = check_if_var_is_invalid(fbot)
    if ftop_invalid and fbot_invalid:
        logger.error("- values for ftop and fbot are invalid!")
        return np.nan
    elif fbot_invalid:
        lay_ftop = get_model_layer_z(ftop, zvec, left=left, right=right)
        logger.error(f"- fbot invalid. selected based on ftop: {lay_ftop}")
        return lay_ftop
    elif ftop_invalid:
        lay_fbot = get_model_layer_z(fbot, zvec, left=left, right=right)
        logger.error(f"- ftop invalid. selected based on fbot: {lay_fbot}")
        return lay_fbot

    if ftop < fbot:
        logger.warning("- tube screen top below tube screen bot, switch top and bot")
        fbot, ftop = ftop, fbot

    lay_ftop = get_model_layer_z(ftop, zvec, left=left, right=right)
    lay_fbot = get_model_layer_z(fbot, zvec, left=left, right=right)

    # Piezometer in layer in which majority of screen is located
    if lay_fbot == lay_ftop:
        logger.debug(f"- selected layer {lay_fbot}")
        return lay_fbot

    else:
        if lay_fbot == left and lay_ftop == right:
            logger.debug("- tube screen spans all layers. " "return nan")
            return np.nan
        elif lay_ftop == right:
            logger.debug(
                "- tube screen top higher than top layer. " f"selected layer {lay_fbot}"
            )
            return lay_fbot

        elif lay_fbot == left:
            logger.debug(
                "- tube screen bot lower than bottom layer. "
                f"selected layer {lay_ftop}"
            )
            return lay_ftop

        logger.debug(
            "- tube screen crosses layer boundary:\n"
            f"  - layers: {lay_ftop}, {lay_fbot}"
        )

        logger.debug(
            f"- tube screen spans {lay_fbot - lay_ftop +1} layers."
            " checking length per layer\n"
            "  - length per layer:"
        )

        # check which layer has the biggest length of the tube screen
        length_layers = np.zeros(lay_fbot - lay_ftop + 1)
        for i in range(len(length_layers)):
            if i == 0:
                length_layers[i] = ftop - zvec[lay_ftop + 1]
            elif (i + 1) == len(length_layers):
                length_layers[i] = zvec[lay_ftop + i] - fbot
            else:
                length_layers[i] = zvec[lay_ftop + i] - zvec[lay_ftop + 1 + i]

            logger.debug(f"    - lay {lay_ftop+i}: {length_layers[i]:.2f}")

        # choose layer with biggest length
        rel_layer = np.argmax(length_layers)
        lay_out = lay_ftop + rel_layer
        logger.debug(f"  - selected layer: {lay_out}")
        return lay_out

    return np.nan


def get_zvec(x, y, gwf=None, ds=None):
    """get a list with the vertical layer boundaries at a point in the model.

    Parameters
    ----------
    x : int or float
        x coordinate.
    y : int or float
        y coordinate.
    gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
        modflow model with top and bottoms
    ds : xarray.Dataset
        xarray Dataset typically create in nlmod. Must have
        dimensions 'x' and 'y' and variables 'top' and 'botm'.

    Raises
    ------
    NotImplementedError
        not all grid types are supported yet.

    Returns
    -------
    zvec : list
        list of vertical layer boundaries. length is nlay + 1.
    """
    import flopy
    from shapely.geometry import Point

    if gwf and not ds:
        ix = flopy.utils.GridIntersect(gwf.modelgrid)
        if gwf.modelgrid.grid_type == "structured":
            res = ix.intersect(Point(x, y))
            if len(res) > 0:
                r, c = res["cellids"][0]
                zvec = np.array(
                    [gwf.modelgrid.top[r, c]]
                    + [gwf.modelgrid.botm[i, r, c] for i in range(gwf.modelgrid.nlay)]
                )
            else:
                print("Point is not within model extent! Returning NaN.")
                zvec = np.nan
        elif gwf.modelgrid.grid_type == "vertex":
            res = ix.intersect([(x, y)], shapetype="point")
            idx = res["cellids"].astype(int)[0]
            if len(res) > 0:
                zvec = [gwf.modelgrid.top[idx]] + [
                    gwf.modelgrid.botm[i, idx] for i in range(gwf.modelgrid.nlay)
                ]
            else:
                print("Point is not within model extent! Returning NaN.")
                zvec = np.nan
        else:
            raise NotImplementedError(
                f"gridtype {gwf.modelgrid.grid_type} not (yet) implemented"
            )
    elif ds and not gwf:
        # assuming modflow type definition with 1 top and N botms
        # check extent
        xmin, xmax, ymin, ymax = ds.attrs["extent"]
        if (x < xmin) or (x > xmax) or (y < ymin) or (y > ymax):
            zvec = np.nan
        elif ds.gridtype == "vertex":
            import nlmod

            cid = nlmod.dims.xy_to_icell2d((x, y), ds)
            sel = ds.sel(icell2d=cid)
            zvec = np.concatenate(([sel["top"].data], sel["botm"].data))
            mask = np.isnan(zvec)
            idx = np.where(~mask, np.arange(mask.size), 0)
            np.maximum.accumulate(idx, axis=0, out=idx)
            zvec[mask] = zvec[idx[mask]]
        else:
            sel = ds.sel(x=x, y=y, method="nearest")
            first_notna = np.nonzero(np.isfinite(sel["top"].data))[0][0]
            zvec = np.concatenate([sel["top"].data[[first_notna]], sel["botm"].data])
            mask = np.isnan(zvec)
            idx = np.where(~mask, np.arange(mask.size), 0)
            np.maximum.accumulate(idx, axis=0, out=idx)
            zvec[mask] = zvec[idx[mask]]
    else:
        raise ValueError("Pass one of 'gwf' or 'ds'!")

    return zvec


@accessor.register_obscollection_accessor("gwobs")
class GwObsAccessor:
    def __init__(self, oc_obj):
        self._obj = oc_obj

    def set_tube_nr(
        self, radius=1, xcol="x", ycol="y", if_exists="error", add_to_meta=False
    ):
        """This method computes the tube numbers based on the location of the
        observations.

        Then it sets the value of the tube number:

        - in the ObsCollection dataframe
        - as the attribute of an Obs object
        - in the meta dictionary of the Obs object (only if add_to_meta is
          True)


        This method is useful for groundwater observations. If two or more
        observation points are close to each other they will be seen as one
        monitoring_well with multiple tubes. The tube_nr is based on the
        'screen_bottom' attribute of the observations in such a way that
        the deepest tube has the highest tube number.

        Parameters
        ----------
        radius : int, optional
            max distance between two observations to be seen as one location,
            by default 1
        xcol : str, optional
            column name with x coordinates, by default 'x'
        ycol : str, optional
            column name with y coordinates, by default 'y'
        if_exists : str, optional
            what to do if an observation point already has a tube_nr, options:
            'error', 'replace' or 'keep', by default 'error'
        add_to_meta : bool, optional
            if True the tube_nr is added to the meta dictionary
            of an observation. The default is False.

        Raises
        ------
        RuntimeError
            if the column tube_nr exists and if_exists='error' an error is
            raised
        """

        # check if column exists in obscollection
        if "tube_nr" in self._obj.columns:
            # set type to numeric
            if self._obj["tube_nr"].dtype != np.number:
                self._obj["tube_nr"] = pd.to_numeric(
                    self._obj["tube_nr"], errors="coerce"
                )

            # check if name should be replaced
            if if_exists == "error":
                raise RuntimeError(
                    "the column 'tube_nr' already exist, set"
                    "if_exists='replace' to replace the current values"
                )
            elif if_exists == "replace":
                self._obj["tube_nr"] = np.nan
        else:
            self._obj["tube_nr"] = np.nan

        # apply tube numbers to tubes that are close to eachother
        for name in self._obj.index:
            if np.isnan(self._obj.loc[name, "tube_nr"]):
                x = self._obj.loc[name, xcol]
                y = self._obj.loc[name, ycol]
                distance_to_other_tubes = np.sqrt(
                    (self._obj[xcol] - x) ** 2 + (self._obj[ycol] - y) ** 2
                )
                dup_x = self._obj.loc[distance_to_other_tubes < radius]
                if dup_x.shape[0] == 1:
                    self._obj._set_metadata_value(
                        name, "tube_nr", 1, add_to_meta=add_to_meta
                    )
                else:
                    dup_x2 = dup_x.sort_values("screen_bottom", ascending=False)
                    for i, pb_dub in enumerate(dup_x2.index):
                        self._obj._set_metadata_value(
                            pb_dub, "tube_nr", i + 1, add_to_meta=add_to_meta
                        )

    def set_tube_nr_monitoring_well(
        self,
        loc_col,
        radius=1,
        xcol="x",
        ycol="y",
        if_exists="error",
        add_to_meta=False,
    ):
        """This method sets the tube_nr and monitoring_well name of an observation
        point based on the location of the observations.

        When two or more tubes are close to another, as defined by radius,
        they are set to the same `monitoring_well` and an increasing `tube_nr` based
        on depth.

        The value of the tube_nr and the monitoring_well are set:

        - in the ObsCollection dataframe
        - as the attribute of an Obs object
        - in the meta dictionary of the Obs object (only if add_to_meta is
          True)

        This method is useful for groundwater observations. If two or more
        observation points are close to each other they will be seen as one
        monitoring_well with multiple tubes. The tube_nr is based on the
        'screen_bottom' attribute of the observations in such a way that
        the deepest tube has the highest tube number. The monitoring_well is
        based on the named of the loc_col of the screen with the lowest
        tube_nr.

        Parameters
        ----------
        loc_col : str
            the column name with the names to use for the monitoring_well
        radius : int, optional
            max distance between two observations to be seen as one location,
            by default 1
        xcol : str, optional
            column name with x coordinates, by default 'x'
        ycol : str, optional
            column name with y coordinates, by default 'y'
        if_exists : str, optional
            what to do if an observation point already has a tube_nr, options:
            'error', 'replace' or 'keep', by default 'error'
        add_to_meta : bool, optional
            if True the tube_nr and location are added to the meta dictionary
            of an observation. The default is False.

        Raises
        ------
        RuntimeError
            if the column tube_nr exists and if_exists='error' an error
            is raised
        """

        # check if columns exists in obscollection
        if "tube_nr" in self._obj.columns or "monitoring_well" in self._obj.columns:
            if if_exists == "error":
                raise RuntimeError(
                    "the column 'tube_nr or monitoring_well' already exist, set"
                    "if_exists='replace' to replace the current values"
                )
            elif if_exists == "replace":
                self._obj["tube_nr"] = np.nan
                self._obj["monitoring_well"] = np.nan
        else:
            self._obj["tube_nr"] = np.nan
            self._obj["monitoring_well"] = np.nan

        # apply tube numbers to tubes that are close to eachother
        for name in self._obj.index:
            if np.isnan(self._obj.loc[name, "tube_nr"]):
                x = self._obj.loc[name, xcol]
                y = self._obj.loc[name, ycol]
                distance_to_other_tubes = np.sqrt(
                    (self._obj[xcol] - x) ** 2 + (self._obj[ycol] - y) ** 2
                )
                dup_x = self._obj.loc[distance_to_other_tubes < radius]
                if dup_x.shape[0] == 1:
                    self._obj._set_metadata_value(
                        name, "tube_nr", 1, add_to_meta=add_to_meta
                    )
                    monitoring_well = self._obj.loc[name, loc_col]
                    self._obj._set_metadata_value(
                        name,
                        "monitoring_well",
                        monitoring_well,
                        add_to_meta=add_to_meta,
                    )
                else:
                    dup_x2 = dup_x.sort_values("screen_bottom", ascending=False)
                    for i, pb_dub in enumerate(dup_x2.index):
                        if i == 0:
                            monitoring_well = self._obj.loc[pb_dub, loc_col]
                        self._obj._set_metadata_value(
                            pb_dub, "tube_nr", i + 1, add_to_meta=add_to_meta
                        )
                        self._obj._set_metadata_value(
                            pb_dub,
                            "monitoring_well",
                            monitoring_well,
                            add_to_meta=add_to_meta,
                        )

    def get_modellayers(self, gwf=None, ds=None):
        """Get the modellayer per observation. The layers can be obtained from
        the modflow model or can be defined in zgr.

        Parameters
        ----------
        gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
            modflow model
        ds : xarray.Dataset
            xarray Dataset with with top and bottoms, must have
            dimensions 'x' and 'y' and variables 'top' and 'bot'

        Returns
        -------
        pd.Series with the modellayers of each observation
        """
        modellayers = []
        for o in self._obj.obs.values:
            logger.debug("-" * 10 + f"\n {o.name}:")

            modellayers.append(o.gwobs.get_modellayer_modflow(gwf=gwf, ds=ds))

        modellayers = pd.Series(
            index=self._obj.index, data=modellayers, name="modellayer"
        )

        return modellayers

    def get_regis_layers(self):
        """Get the regis layer per observation.

        Parameters
        ----------


        Returns
        -------
        pd.Series with the names of the regis layer of each observation
        """

        return self._obj.obs.apply(lambda o: o.gwobs.get_regis_layer())


@accessor.register_obs_accessor("gwobs")
class GeoAccessorObs:
    def __init__(self, obs):
        self._obj = obs

    def get_modellayer_modflow(self, gwf=None, ds=None, left=-999, right=999):
        """Add modellayer to meta dictionary.

        Parameters
        ----------
        gwf : flopy.mf6.modflow.mfgwf.ModflowGwf
            modflow model
        ds : xarray.Dataset
            xarray Dataset with with top and bottoms, must have
            dimensions 'x' and 'y' and variables 'top' and 'bot'

        Returns
        -------
        int
            modellayer
        """

        zvec = get_zvec(self._obj.x, self._obj.y, gwf=gwf, ds=ds)

        if np.all(np.isnan(zvec)):
            return np.nan
        else:
            modellayer = get_modellayer_from_screen_depth(
                self._obj.screen_top,
                self._obj.screen_bottom,
                zvec,
                left=left,
                right=right,
            )
            return modellayer

    def get_regis_layer(self):
        """find the name of the REGIS layer based on the tube screen depth.

        Parameters
        ----------


        Returns
        -------
        str
            name of REGIS layer
        """
        import xarray as xr

        if np.isnan(self._obj.screen_top) or np.isnan(self._obj.screen_bottom):
            return "nan"

        # connect to regis netcdf
        regis_url = r"http://www.dinodata.nl:80/opendap/REGIS/REGIS.nc"
        regis_ds = xr.open_dataset(regis_url, decode_times=False)

        # rename layer in regis netcdf
        regis_ds = regis_ds.rename({"layer": "layer_old"})
        regis_ds.coords["layer"] = regis_ds.layer_old.astype(str)
        regis_ds = regis_ds.swap_dims({"layer_old": "layer"})

        # get zvec regis netcdf
        z = (
            regis_ds["bottom"]
            .sel(x=self._obj.x, y=self._obj.y, method="nearest")
            .dropna(dim="layer")
        )
        ztop = regis_ds["top"].sel(x=self._obj.x, y=self._obj.y, method="nearest").max()
        zvec = np.concatenate([[ztop.values], z.values])

        # get index of regis model layer
        layer_i = get_modellayer_from_screen_depth(
            self._obj.screen_top,
            self._obj.screen_bottom,
            zvec,
            left=-999,
            right=999,
        )

        if layer_i == 999:
            return "above"
        elif layer_i == -999:
            return "below"

        return str(z.layer[layer_i].values)



# -*- coding: utf-8 -*-
"""Created on Fri Nov  8 11:46:51 2019

@author: oebbe
"""

import test_001_to_from as ttf

plot_dir = "./tests/data/2019-Dino-test/plots"


def test_interactive_plot():
    gw = ttf.test_observation_gw()
    gw.plots.interactive_plot(
        savedir=plot_dir,
        cols=("stand_m_tov_nap",),
        hoover_date_format="{%F}",
        add_screen_to_legend=True,
    )
    return


def test_obscollection_dino_to_imap():
    dino_gw = ttf.test_obscollection_dinozip_gw()
    dino_gw.geo.set_lat_lon()
    dino_gw.plots.interactive_map(
        plot_dir,
        cols=("stand_m_tov_nap",),
        fname="imap.html",
        legend_name="grondwater DINO",
        add_screen_to_legend=True,
        hoover_names=("gws",),
        zoom_start=9,
    )
    return


def test_obscollection_dino_to_mapgraph():
    try:
        from art_tools import hpd_extension  # noqa: F401

        gw = ttf.test_obscollection_dinozip_gw()
        gw.art.plot_mapgraphs(plot_ylim="min_dy")
        return
    except ModuleNotFoundError as e:
        print(e)
        return


def test_obscollection_dino_to_section_plot():
    dino_gw = ttf.test_obscollection_dinozip_gw()
    dino_gw.plots.section_plot()
    return


def test_obscollection_to_map():
    try:
        from art_tools import hpd_extension  # noqa: F401

        fews_gw_prod = ttf.test_obscollection_fews_lowmemory()
        fews_gw_prod.art.plot_mapfig()
        return
    except ModuleNotFoundError as e:
        print(e)
        return


def test_obscollection_to_imap():
    fname = "texel_fews.html"
    fews_gw_prod = ttf.test_obscollection_fews_lowmemory()
    # add metadata to obscollection DF
    fews_gw_prod.add_meta_to_df("lat")
    fews_gw_prod.add_meta_to_df("lon")
    # convert columns to float
    fews_gw_prod["lat"] = fews_gw_prod["lat"].astype(float)
    fews_gw_prod["lon"] = fews_gw_prod["lon"].astype(float)

    fews_gw_prod.gwobs.set_tube_nr_monitoring_well(
        "monitoring_well", if_exists="replace"
    )

    fews_gw_prod.plots.interactive_map(
        plot_dir,
        plot_columns=("value",),
        fname=fname,
        plot_freq="D",
        legend_name="opp water FEWS",
        map_label="locationId",
        map_label_size=10,
    )
    return



"""
https://www.bro-productomgeving.nl/bpo/latest/informatie-voor-softwareleveranciers/url-s-publieke-rest-services
"""

import json
import logging
import xml.etree.ElementTree

import numpy as np
import pandas as pd
import requests
from pyproj import Transformer
from tqdm import tqdm

logger = logging.getLogger(__name__)


def get_obs_list_from_gmn(bro_id, ObsClass, only_metadata=False, keep_all_obs=True):
    """get a list of observation from a groundwater monitoring network.

    Parameters
    ----------
    bro_id : str
        starts with 'GMN' e.g. 'GMN000000000163'.
    ObsClass : type
        class of the observations, so far only GroundwaterObs is supported
    only_metadata : bool, optional
        if True download only metadata, significantly faster. The default
        is False.
    keep_all_obs : boolean, optional
        add all observation points to the collection, even without
        measurements

    Raises
    ------
    ValueError
        DESCRIPTION.

    Returns
    -------
    obs_list : list
        list with observation objects.
    meta : dict
        metadata of the groundwater monitoring net.

    """

    if not bro_id.startswith("GMN"):
        raise ValueError("bro id should start with GMN")

    url = f"https://publiek.broservices.nl/gm/gmn/v1/objects/{bro_id}"
    req = requests.get(url)

    if req.status_code > 200:
        print(req.json()["errors"][0]["message"])

    ns = {"xmlns": "http://www.broservices.nl/xsd/dsgmn/1.0"}

    tree = xml.etree.ElementTree.fromstring(req.text)
    gmn = tree.find(".//xmlns:GMN_PO", ns)
    gmws = gmn.findall("xmlns:measuringPoint", ns)

    logger.info(
        f"{len(gmws)} groundwater monitoring wells within groundwater meetnet")

    obs_list = []
    for gmw in tqdm(gmws):
        tags = ["MeasuringPoint", "monitoringTube", "GroundwaterMonitoringTube"]
        tube = gmw.find(f".//xmlns:{'//xmlns:'.join(tags)}", ns)
        gmw_id = tube.find("xmlns:broId", ns).text
        tube_nr = int(tube.find("xmlns:tubeNumber", ns).text)

        o = ObsClass.from_bro(
            bro_id=gmw_id, tube_nr=tube_nr, only_metadata=only_metadata
        )
        if o.empty:
            logger.warning(
                f"no measurements found for gmw_id {gmw_id} and tube number {tube_nr}"
            )
            if keep_all_obs:
                obs_list.append(o)
        else:
            obs_list.append(o)
        obs_list.append(o)

    meta = {}
    meta["name"] = gmn.find("xmlns:name", ns).text
    meta["doel"] = gmn.find("xmlns:monitoringPurpose", ns).text
    meta["bro_id"] = bro_id

    return obs_list, meta


def get_bro_groundwater(bro_id, tube_nr=None, only_metadata=False, **kwargs):
    """get bro groundwater measurement from a GLD id or a GMW id with a
    filter number.


    Parameters
    ----------
    bro_id : str
        starts with 'GLD' or 'GMW' e.g. 'GLD000000012893'.
    tube_nr : int or None, optional
        tube number, required if bro_id starts with 'GMW'. The default is
        None.
    only_metadata : bool, optional
        if True download only metadata, significantly faster. The default
        is False.
    **kwargs :
        passes to measurements_from_gld.

    Raises
    ------
    ValueError
        DESCRIPTION.

    Returns
    -------
    dataframe
        measurements.
    dictionary
        metadata.

    """
    logger.info(f"reading bro_id {bro_id}")

    if bro_id.startswith("GLD"):
        if only_metadata:
            raise ValueError("cannot get metadata from gld id")
        return measurements_from_gld(bro_id, **kwargs)

    elif bro_id.startswith("GMW"):
        if tube_nr is None:
            raise ValueError(
                "if bro_id is GMW a filternumber should be specified")

        meta = get_metadata_from_gmw(bro_id, tube_nr)
        gld_ids = get_gld_ids_from_gmw(bro_id, tube_nr)

        if gld_ids is None:
            meta["name"] = f"{bro_id}_{tube_nr}"
            only_metadata = True  # cannot get time series without gld id
        else:
            meta["name"] = f"{bro_id}_{tube_nr}"
            meta["gld_ids"] = gld_ids

        if only_metadata:
            empty_df = pd.DataFrame()
            return empty_df, meta

        for i, gld_id in enumerate(gld_ids):
            if i == 0:
                df, meta_new = measurements_from_gld(gld_id, **kwargs)
                meta.update(meta_new)
            else:
                df_new, meta_new = measurements_from_gld(gld_id, **kwargs)
                df = pd.concat([df, df_new], axis=1)
                meta.update(meta_new)

        return df, meta


def get_gld_ids_from_gmw(bro_id, tube_nr):
    """get bro_ids of multiple grondwaterstandendossier (gld) from a bro_id of a
    grondwatermonitoringsput (gmw).

    Parameters
    ----------
    bro_id : str
        starts with 'GLD' or 'GMW' e.g. 'GMW000000036287'.
    tube_nr : int
        tube number.

    Raises
    ------
    ValueError
        DESCRIPTION.
    RuntimeError
        DESCRIPTION.

    Returns
    -------
    list of str or None
        bro_ids of a grondwaterstandonderzoek (gld).

    """
    if not bro_id.startswith("GMW"):
        raise ValueError("bro id should start with GMW")

    url = f"https://publiek.broservices.nl/gm/v1/gmw-relations/{bro_id}"
    req = requests.get(url)

    if req.status_code > 200:
        print(req.json()["errors"][0]["message"])

    d = json.loads(req.text)

    if len(d["monitoringTubeReferences"]) == 0:
        logger.info(
            f"no groundwater level dossier for {bro_id} and tube number {tube_nr}"
        )
        return None

    for tube in d["monitoringTubeReferences"]:
        if tube["tubeNumber"] == tube_nr:
            if len(tube["gldReferences"]) == 0:
                logger.info(
                    f"no groundwater level dossier for {bro_id} and tube number"
                    f"{tube_nr}"
                )
                return None
            else:
                return [gldref["broId"] for gldref in tube["gldReferences"]]


def measurements_from_gld(
    bro_id, tmin=None, tmax=None, to_wintertime=True, drop_duplicate_times=True
):
    """get measurements and metadata from a grondwaterstandonderzoek (gld)
    bro_id


    Parameters
    ----------
    bro_id : str
        e.g. 'GLD000000012893'.
    tmin : str or None, optional
        start date in format YYYY-MM-DD
    tmax : str or None, optional
        end date in format YYYY-MM-DD
    to_wintertime : bool, optional
        if True the time index is converted to Dutch winter time. The default
        is True.
    drop_duplicate_times : bool, optional
        if True rows with a duplicate time stamp are removed keeping only the
        first row. The default is True.
    add_registration_history : bool, optional
        if True the registration history is added to the metadata. The defualt
        is True.

    Raises
    ------
    ValueError
        if bro_id is invalid.

    Returns
    -------
    df : pd.DataFrame
        measurements.
    meta : dict
        metadata.

    """
    if not bro_id.startswith("GLD"):
        raise ValueError("can only get observations if bro id starts with GLD")

    url = "https://publiek.broservices.nl/gm/gld/v1/objects/{}"
    params = {}
    if tmin is not None:
        tmin = pd.to_datetime(tmin)
        params["observationPeriodBeginDate"] = tmin.strftime("%Y-%m-%d")
    if tmax is not None:
        tmax = pd.to_datetime(tmax)
        params["observationPeriodEndDate"] = tmax.strftime("%Y-%m-%d")
    req = requests.get(url.format(bro_id), params=params)

    if req.status_code > 200:
        print(req.json()["errors"][0]["message"])

    ns = {
        "ns11": "http://www.broservices.nl/xsd/dsgld/1.0",
        "gldcommon": "http://www.broservices.nl/xsd/gldcommon/1.0",
        "waterml": "http://www.opengis.net/waterml/2.0",
        "swe": "http://www.opengis.net/swe/2.0",
        "om": "http://www.opengis.net/om/2.0",
    }

    tree = xml.etree.ElementTree.fromstring(req.text)

    glds = tree.findall(".//ns11:GLD_O", ns)
    if len(glds) != 1:
        raise (Exception("Only one gld supported"))
    gld = glds[0]

    meta = {"source": "BRO"}
    meta["monitoring_well"] = gld.find(
        "ns11:monitoringPoint//gldcommon:broId", ns).text
    meta["tube_nr"] = int(
        gld.find("ns11:monitoringPoint//gldcommon:tubeNumber", ns).text
    )
    meta["name"] = f"{meta['monitoring_well']}_{meta['tube_nr']}"
    gmn = gld.find("ns11:groundwaterMonitoringNet//gldcommon:broId", ns)
    if gmn is None:
        meta["monitoringsnet"] = None
    else:
        meta["monitoringsnet"] = gmn.text

    # get observations
    msts = "ns11:observation//om:result//waterml:MeasurementTimeseries"
    times = [time.text for time in gld.findall(f"{msts}//waterml:time", ns)]
    values = [
        np.nan if value.text is None else float(value.text)
        for value in gld.findall(f"{msts}//waterml:value", ns)
    ]
    qualifiers = [q.text for q in gld.findall(
        f"{msts}//swe:Category//swe:value", ns)]

    # to dataframe
    df = pd.DataFrame(
        index=pd.to_datetime(times),
        data={"values": values, "qualifier": qualifiers},
    )

    # wintertime
    if to_wintertime:
        # remove time zone information by transforming to dutch winter time
        df.index = pd.to_datetime(df.index, utc=True).tz_localize(None) + pd.Timedelta(
            1, unit="H"
        )

    # duplicates
    if df.index.has_duplicates and drop_duplicate_times:
        duplicates = df.index.duplicated(keep="first")
        message = "{} contains {} duplicates (of {}). Keeping only first values."
        logger.info(message.format(bro_id, duplicates.sum(), len(df)))
        df = df[~duplicates]

    df = df.sort_index()

    # slice to tmin and tmax
    df = df.loc[tmin:tmax]

    # add metadata from gmw
    meta.update(get_metadata_from_gmw(meta["monitoring_well"], meta["tube_nr"]))

    return df, meta


def get_full_metadata_from_gmw(bro_id, tube_nr):
    """get metadata for a groundwater monitoring well.


    Parameters
    ----------
    bro_id : str
        bro id of groundwater monitoring well e.g. 'GMW000000036287'.
    tube_nr : int
        filter number you want metadata for.

    Raises
    ------
    ValueError
        if bro_id is invalid.

    Returns
    -------
    meta : dict
        dictionary with metadata.

    """

    if not bro_id.startswith("GMW"):
        raise ValueError("can only get metadata if bro id starts with GMW")

    url = f"https://publiek.broservices.nl/gm/gmw/v1/objects/{bro_id}"
    req = requests.get(url)

    # read results
    tree = xml.etree.ElementTree.fromstring(req.text)
    ns = "{http://www.broservices.nl/xsd/dsgmw/1.1}"

    gmws = tree.findall(f".//{ns}GMW_PO")
    if len(gmws) != 1:
        raise (Exception("Only one gmw supported"))
    gmw = gmws[0]
    meta = {"monitoring_well": bro_id, "tube_nr": tube_nr, "source": "BRO"}
    for child in gmw:
        key = child.tag.split("}", 1)[1]
        if len(child) == 0:
            meta[key] = child.text
        elif key == "deliveredLocation":
            ns = "{http://www.broservices.nl/xsd/gmwcommon/1.1}"
            point = child.find(f"{ns}location")
            ns = "{http://www.opengis.net/gml/3.2}"
            xy = [float(x) for x in point.find(f"{ns}pos").text.split()]
            meta["x"], meta["y"] = xy
        elif key in ["wellConstructionDate"]:
            meta[key] = child[0].text
        elif key == "wellHistory":
            for grandchild in child:
                key = grandchild.tag.split("}", 1)[1]
                meta[key] = grandchild[0].text
        elif key in ["deliveredVerticalPosition", "registrationHistory"]:
            for grandchild in child:
                key = grandchild.tag.split("}", 1)[1]
                meta[key] = grandchild.text
        elif key in ["monitoringTube"]:
            tube = False
            tube_dic = {}
            for grandchild in child:
                if len(grandchild) == 0:
                    tube_key = grandchild.tag.split("}", 1)[1]
                    if tube_key == "tubeNumber":
                        if int(grandchild.text) == tube_nr:
                            tube = True
                    else:
                        tube_dic[tube_key] = grandchild.text

                else:
                    for greatgrandchild in grandchild:
                        tube_key = greatgrandchild.tag.split("}", 1)[1]
                        tube_dic[tube_key] = greatgrandchild.text
            if tube:
                meta.update(tube_dic)

    rename_dic = {
        "groundLevelPosition": "ground_level",
        "screenTopPosition": "screen_top",
        "screenBottomPosition": "screen_bottom",
        "tubeTopPosition": "tube_top",
    }

    for key, val in rename_dic.items():
        meta[val] = meta.pop(key)

    return meta


def get_metadata_from_gmw(bro_id, tube_nr):
    """get selection of metadata for a groundwater monitoring well.
    coordinates, ground_level, tube_top and tube screen


    Parameters
    ----------
    bro_id : str
        bro id of groundwater monitoring well e.g. 'GMW000000036287'.
    tube_nr : int
        tube number you want metadata for.

    Raises
    ------
    ValueError
        if bro_id is invalid.
    TypeError
        if tube_nr is not an int

    Returns
    -------
    meta : dict
        dictionary with metadata.

    """
    ns = {
        "dsgmw": "http://www.broservices.nl/xsd/dsgmw/1.1",
        "gmwcommon": "http://www.broservices.nl/xsd/gmwcommon/1.1",
        "gml": "http://www.opengis.net/gml/3.2",
    }

    if not bro_id.startswith("GMW"):
        raise ValueError("can only get metadata if bro id starts with GMW")

    if not isinstance(tube_nr, int):
        raise TypeError(f"expected integer got {type(tube_nr)}")

    url = f"https://publiek.broservices.nl/gm/gmw/v1/objects/{bro_id}"
    req = requests.get(url)

    # read results
    tree = xml.etree.ElementTree.fromstring(req.text)

    gmws = tree.findall(".//dsgmw:GMW_PO", ns)
    if len(gmws) != 1:
        raise (Exception("Only one gmw supported"))
    gmw = gmws[0]

    meta = {"monitoring_well": bro_id, "tube_nr": tube_nr, "source": "BRO"}

    # x and y
    xy = gmw.find("dsgmw:deliveredLocation//gmwcommon:location//gml:pos", ns)
    meta["x"], meta["y"] = [float(val) for val in xy.text.split()]

    # ground_level
    vert_pos = gmw.find("dsgmw:deliveredVerticalPosition", ns)
    mv = vert_pos.find("gmwcommon:groundLevelPosition", ns)
    datum = vert_pos.find("gmwcommon:verticalDatum", ns)
    if datum.text == "NAP" and mv.attrib["uom"] == "m":
        meta["unit"] = "m NAP"
        if mv.text is None:
            meta["ground_level"] = np.nan
        else:
            meta["ground_level"] = float(mv.text)
    else:
        raise ValueError("invalid ground_level datum or unit")

    # buis eigenschappen
    tubes = gmw.findall("dsgmw:monitoringTube", ns)
    for tube in tubes:
        if int(tube.find("dsgmw:tubeNumber", ns).text) == tube_nr:
            break

    # tube_top
    mp = tube.find("dsgmw:tubeTopPosition", ns)
    if mp.attrib["uom"] == "m":
        meta["tube_top"] = float(mp.text)

    # bovenkant filter
    bkf = tube.find("dsgmw:screen//dsgmw:screenTopPosition", ns)
    if bkf.attrib["uom"] == "m":
        meta["screen_top"] = float(bkf.text)

    # onderkant filter
    okf = tube.find("dsgmw:screen//dsgmw:screenBottomPosition", ns)
    if okf.attrib["uom"] == "m":
        meta["screen_bottom"] = float(okf.text)

    meta["metadata_available"] = True

    return meta


def get_obs_list_from_extent(
    extent,
    ObsClass,
    tmin=None,
    tmax=None,
    only_metadata=False,
    keep_all_obs=True,
    epsg=28992,
    ignore_max_obs=False,
):
    """get a list of gmw observations within an extent.


    Parameters
    ----------
    extent : list, tuple, numpy-array or None, optional
        get groundwater monitoring wells within this extent
        [xmin, xmax, ymin, ymax]
    ObsClass : type
        class of the observations, e.g. GroundwaterObs or WaterlvlObs
    tmin : str or None, optional
        start time of observations. The default is None.
    tmax : str or None, optional
        end time of observations. The default is None.
    only_metadata : bool, optional
        if True download only metadata, significantly faster. The default
        is False.
    epsg : int, optional
        epsg code of the extent. The default is 28992 (RD).
    ignore_max_obs : bool, optional
        by default you get a prompt if you want to download over a 1000
        observations at once. if ignore_max_obs is True you won't get the
        prompt. The default is False

    Raises
    ------

        DESCRIPTION.

    Returns
    -------
    obs_list : TYPE
        DESCRIPTION.

    """

    if only_metadata and not keep_all_obs:
        logger.error(
            "you will get an empty ObsCollection with only_metadata is True and"
            "keep_all_obs is False"
        )

    url = "https://publiek.broservices.nl/gm/gmw/v1/characteristics/searches?"

    data = {}
    if tmin is None or tmax is None:
        data["registrationPeriod"] = {}
        if tmin is not None:
            beginDate = pd.to_datetime(tmin).strftime("%Y-%m-%d")
            data["registrationPeriod"]["beginDate"] = beginDate
        if tmax is not None:
            endDate = pd.to_datetime(tmax).strftime("%Y-%m-%d")
            data["registrationPeriod"]["endDate"] = endDate

    transformer = Transformer.from_crs(epsg, 4326)
    data["area"] = {}
    if extent is not None:
        lat1, lon1 = transformer.transform(extent[0], extent[2])
        lat2, lon2 = transformer.transform(extent[1], extent[3])
        data["area"]["boundingBox"] = {
            "lowerCorner": {"lat": lat1, "lon": lon1},
            "upperCorner": {"lat": lat2, "lon": lon2},
        }
    req = requests.post(url, json=data)
    if req.status_code > 200:
        print(req.json()["errors"][0]["message"])

    # read results
    tree = xml.etree.ElementTree.fromstring(req.text)

    ns = {
        "dsgmw": "http://www.broservices.nl/xsd/dsgmw/1.1",
        "gml": "http://www.opengis.net/gml/3.2",
        "brocom": "http://www.broservices.nl/xsd/brocommon/3.0",
    }

    if tree.find(".//brocom:responseType", ns).text == "rejection":
        raise RuntimeError(tree.find(".//brocom:rejectionReason", ns).text)

    gmws_ids = np.unique(
        [gmw.text for gmw in tree.findall(".//dsgmw:GMW_C//brocom:broId", ns)]
    )

    if len(gmws_ids) > 1000 and not ignore_max_obs:
        ans = input(
            f"You requested to download {len(gmws_ids)} observations, this can"
            "take a while. Are you sure you want to continue [Y/n]? "
        )
        if ans not in ["Y", "y", "yes", "Yes", "YES"]:
            return []

    obs_list = []
    for gmw_id in tqdm(gmws_ids):
        gmws = tree.findall(f'.//*[brocom:broId="{gmw_id}"]', ns)
        if len(gmws) < 1:
            raise RuntimeError("unexpected")
        else:
            gmw = gmws[0]

        ntubes = int(gmw.find("dsgmw:numberOfMonitoringTubes", ns).text)
        for tube_nr in range(1, ntubes + 1):
            o = ObsClass.from_bro(
                gmw_id,
                tube_nr=tube_nr,
                tmin=tmin,
                tmax=tmax,
                only_metadata=only_metadata,
            )
            if o.empty:
                logger.warning(
                    f"no measurements found for gmw_id {gmw_id} and tube number"
                    f"{tube_nr}"
                )
                if keep_all_obs:
                    obs_list.append(o)
            else:
                obs_list.append(o)

    return obs_list



# -*- coding: utf-8 -*-
"""Created on Wed Sep 12 12:15:42 2018.

@author: Artesia
"""
import logging
import numbers

logger = logging.getLogger(__name__)


def _get_metadata_from_obs(o):
    """internal method to get metadata in the right format for a pastas series.
    A pastas timeseries cannot handle the same metadata format as an
    observation object.

    Parameters
    ----------
    o : observations.Obs
        observation time series with metadata

    Returns
    -------
    meta : dictionary
        meta dictionary.
    """
    meta = dict()
    for attr_key in o._metadata:
        val = getattr(o, attr_key)
        if isinstance(val, (int, float, str, bool)):
            meta[attr_key] = val
        elif isinstance(val, numbers.Number):
            meta[attr_key] = float(val)
        elif isinstance(val, dict):
            for k, v in val.items():
                if isinstance(v, (int, float, str, bool)):
                    meta[k] = v
                elif isinstance(v, numbers.Number):
                    meta[k] = float(v)
                else:
                    logger.info(
                        f"did not add {k} to metadata because datatype is {type(v)}"
                    )
        else:
            logger.info(
                f"did not add {attr_key} to metadata because datatype is {type(val)}"
            )

    return meta


def create_pastastore(
    oc,
    pstore,
    pstore_name="",
    conn=None,
    add_metadata=True,
    col=None,
    kind="oseries",
    overwrite=False,
):
    """add observations to a new or existing pastastore.

    Parameters
    ----------
    oc : observation.ObsCollection
        collection of observations
    pstore : pastastore.PastaStore, optional
            Existing pastastore, if None a new pastastore is created
    pstore_name : str, optional
        Name of the pastastore only used if pstore is None
    conn : pastastore.connectors
        connector for database
    col : str or None, optional
        the column of the obs dataframe to use in pastas. The first numeric
        column is used if col is None, by default None.
    kind : str, optional
        The kind of series that is added to the pastastore
    add_metadata : boolean, optional
        If True metadata from the observations added to the pastastore
    overwrite : boolean, optional
        if True, overwrite existing series in pastastore, default is False

    Returns
    -------
    pstore : pastastore.PastaStore
        the pastastore with the series from the ObsCollection

    """
    import pastastore as pst

    if pstore is None:
        if conn is None:
            conn = pst.DictConnector(name="my_db")
        pstore = pst.PastaStore(name=pstore_name, connector=conn)

    for o in oc.obs.values:
        logger.info("add to pastastore -> {}".format(o.name))

        if add_metadata:
            meta = _get_metadata_from_obs(o)
        else:
            meta = dict()

        if col is None:
            col = o._get_first_numeric_col_name()

        if kind == "oseries":
            pstore.conn.add_oseries(o[col], o.name, metadata=meta, overwrite=overwrite)
        else:
            pstore.conn.add_stress(
                o[col], o.name, kind, metadata=meta, overwrite=overwrite
            )

    return pstore



import logging
import os
import re
import tempfile

import numpy as np
import pandas as pd

from ..util import unzip_file

logger = logging.getLogger(__name__)


def _read_dino_groundwater_header(f):
    line = f.readline()
    header = dict()
    while line not in ["\n", "", "\r\n"]:
        if "," in line:  # for csv from dinoloket
            propval = line.split(",")
            prop = propval[0]
            prop = prop.replace(":", "")
            prop = prop.strip()
            val = propval[1]
            if propval[2] != "":
                val = val + " " + propval[2].replace(":", "") + " " + propval[3]
            header[prop] = val
        else:  # for artdiver dino-csv
            propval = line.split(":")
            prop = propval[0]
            val = ":".join(propval[1:])
            header[prop] = val
        line = f.readline()

    return line, header


def _read_empty(f, line):
    while (line == "\n") or (line == "\r\n"):
        line = f.readline()
    return line


def _read_dino_groundwater_referencelvl(f, line):
    ref = {}
    while line not in ["\n", "", "\r\n"]:
        propval = line.split(",")
        prop = propval[0]
        prop = prop.replace(":", "")
        prop = prop.strip()
        if len(propval) > 1:
            val = propval[1]
            ref[prop] = val
        line = f.readline()
    return line, ref


def _read_dino_groundwater_metadata(f, line):
    _translate_dic_float = {
        "x-coordinaat": "x",
        "y-coordinaat": "y",
        "filternummer": "tube_nr",
    }
    _translate_dic_div_100 = {
        "meetpunt (cm t.o.v. nap)": "tube_top",
        "bovenkant filter (cm t.o.v. nap)": "screen_top",
        "onderkant filter (cm t.o.v. nap)": "screen_bottom",
        "maaiveld (cm t.o.v. nap)": "ground_level",
    }
    metalist = list()
    line = line.strip()
    properties = line.split(",")
    line = f.readline()

    while line not in ["\n", "", "\r\n"]:
        meta = dict()
        line = line.strip()
        values = line.split(",")
        for i, val in enumerate(values):
            meta[properties[i].lower()] = val
        metalist.append(meta)
        line = f.readline()

    meta_ts = {}
    if metalist:
        # add time dependent metadata to meta_ts
        for i, meta in enumerate(metalist):
            meta_tsi = {}
            start_date = pd.to_datetime(meta.pop("startdatum"), dayfirst=True)
            # end_date = pd.to_datetime(meta.pop('einddatum'), dayfirst=True)
            meta_tsi["monitoring_well"] = meta["locatie"]
            for key in _translate_dic_float.keys():
                if meta[key] == "":
                    meta_tsi[_translate_dic_float[key]] = np.nan
                else:
                    meta_tsi[_translate_dic_float[key]] = float(meta[key])

            for key in _translate_dic_div_100.keys():
                if meta[key] == "":
                    meta_tsi[_translate_dic_div_100[key]] = np.nan
                else:
                    meta_tsi[_translate_dic_div_100[key]] = float(meta[key]) / 100.0
            if i == 0:
                for key in meta_tsi.keys():
                    meta_ts[key] = pd.Series(name=key, dtype=type(meta_tsi[key]))

            for key in meta_tsi.keys():
                meta_ts[key].loc[start_date] = meta_tsi[key]

        # remove series with non time variant metadata from meta_ts
        ts_keys = (
            ["monitoring_well"]
            + list(_translate_dic_float.values())
            + list(_translate_dic_div_100.values())
        )
        for key in ts_keys:
            unique_values = meta_ts[key].unique()
            if len(unique_values) == 1:
                meta_ts.pop(key)

        obs_att = meta_tsi.copy()
        obs_att["name"] = f'{obs_att["monitoring_well"]}-{int(obs_att["tube_nr"]):03d}'
        obs_att["metadata_available"] = True
    else:
        # no metadata
        obs_att = {}
        obs_att["monitoring_well"] = ""
        obs_att["tube_nr"] = np.nan
        obs_att["name"] = "unknown"
        obs_att["x"] = np.nan
        obs_att["y"] = np.nan
        obs_att["tube_top"] = np.nan
        obs_att["ground_level"] = np.nan
        obs_att["screen_top"] = np.nan
        obs_att["screen_bottom"] = np.nan
        obs_att["metadata_available"] = False

    return line, obs_att, meta_ts


def _read_dino_groundwater_measurements(f, line):
    line = line.strip()
    titel = line.split(",")
    while "" in titel:
        titel.remove("")

    if line != "":
        # Validate if titles are valid names
        validator = np.lib._iotools.NameValidator()
        titel = [s.lower() for s in validator(titel)]
        usecols = range(0, len(titel))

        try:
            measurements = pd.read_csv(
                f,
                header=None,
                names=titel,
                parse_dates=["peildatum"],
                index_col="peildatum",
                dayfirst=True,
                usecols=usecols,
            )
        except pd.errors.ParserError:
            # for now the workflow is to remove the files that cannot be read
            # manually.
            measurements = None
    else:
        measurements = None

    return line, measurements


def read_dino_groundwater_quality_txt(fname):
    """Read dino groundwater quality (grondwatersamenstelling) from a dinoloket
    txt file.

    Notes
    -----
    this function has not been tested thoroughly

    Parameters
    ----------
    fname : str
        path to txt file

    Returns
    -------
    measurements : pd.DataFrame
    meta : dict
        dictionary with metadata
    """
    logger.info("reading -> {}".format(os.path.split(fname)[-1]))
    with open(fname, "r") as f:
        # LOCATIE gegevens
        line = f.readline().rstrip("\n")
        assert line == "LOCATIE gegevens"

        strt_locatie = f.tell()
        # determine the number of rows
        nrows = -1  # the header does not count
        line = f.readline()
        while line not in ["\n", ""]:
            nrows += 1
            line = f.readline()
        eind_locatie = f.tell()
        # go back to where we were before
        f.seek(strt_locatie)
        # read the location-information
        locatie = pd.read_csv(f, sep="\t", nrows=nrows)
        # there is always only one location (change if this is not the case)
        assert nrows == 1

        locatie = locatie.squeeze()

        # KWALITEIT gegevens VLOEIBAAR
        f.seek(eind_locatie)
        line = f.readline().rstrip("\n")
        assert line == "KWALITEIT gegevens VLOEIBAAR"

        measurements = pd.read_csv(
            f,
            sep="\t",
            parse_dates=["Monster datum", "Analyse datum"],
            dayfirst=True,
            index_col="Monster datum",
        )

    meta = {
        "filename": fname,
        "source": "dino",
        "monitoring_well": locatie["NITG-nr"],
        "name": locatie["NITG-nr"],
        "x": locatie["X-coord"],
        "y": locatie["Y-coord"],
    }
    try:
        meta["ground_level"] = locatie["Maaiveldhoogte (m tov NAP)"]
    except KeyError:
        meta["ground_level"] = np.nan

    return measurements, meta


def read_dino_groundwater_csv(
    fname, to_mnap=True, read_series=True, remove_duplicates=False, keep_dup="last"
):
    """Read dino groundwater quantity data from a dinoloket csv file.

    Parameters
    ----------
    fname : str
        path to csv file
    to_mnap : boolean, optional
        if True a column with 'stand_m_tov_nap' is added to the dataframe
    read_series : boolean, optional
        if False only metadata is read, default is True
    remove_duplicates : boolean, optional
        if True duplicate indices are removed. Default is False.
    keep_dup : str, optional
        indicate which duplicate indices should be kept, only used when
        remove_duplicates is True. Default is 'last'

    Returns
    -------
    measurements : pd.DataFrame
    meta : dict
        dictionary with metadata
    """
    logger.info(f"reading -> {os.path.split(fname)[-1]}")

    with open(fname, "r") as f:
        # read header
        line, header = _read_dino_groundwater_header(f)
        line = _read_empty(f, line)
        if not header:
            logger.warning(f"could not read header -> {fname}")

        # read reference level
        line, ref = _read_dino_groundwater_referencelvl(f, line)
        line = _read_empty(f, line)
        if not ref:
            logger.warning(f"could not read reference level -> {fname}")

        # read metadata
        line, meta, meta_ts = _read_dino_groundwater_metadata(f, line)
        line = _read_empty(f, line)
        if not meta["metadata_available"]:
            logger.warning(f"could not read metadata -> {fname}")
        meta["filename"] = fname
        meta["source"] = "dino"

        # read measurements
        if read_series:
            line, measurements = _read_dino_groundwater_measurements(f, line)
            if measurements is None:
                logger.warning(f"could not read measurements -> {fname}")
            elif measurements[~measurements.stand_cm_tov_nap.isna()].empty:
                logger.warning(f"no NAP measurements available -> {fname}")
            if to_mnap and measurements is not None:
                measurements.insert(
                    0, "stand_m_tov_nap", measurements["stand_cm_tov_nap"] / 100.0
                )
                meta["unit"] = "m NAP"
            elif not to_mnap:
                meta["unit"] = "cm NAP"
            if remove_duplicates:
                measurements = measurements[
                    ~measurements.index.duplicated(keep=keep_dup)
                ]

            # add time variant metadata to measurements
            for s in meta_ts.values():
                if measurements is None:
                    measurements = pd.DataFrame(data=s.copy(), columns=[s.name])
                else:
                    measurements = measurements.join(s, how="outer")
                measurements.loc[:, s.name] = measurements.loc[:, s.name].ffill()

        else:
            measurements = None

    return measurements, meta


def _read_artdino_groundwater_metadata(f, line):
    metalist = list()
    line = line.strip()
    properties = line.split(",")
    line = f.readline()
    while line not in ["\n", "", "\r\n"]:
        meta = dict()
        line = line.strip()
        values = line.split(",")
        for i in range(0, len(values)):
            meta[properties[i].lower()] = values[i]
        metalist.append(meta)
        line = f.readline()

    meta = {}
    if metalist:
        meta["monitoring_well"] = metalist[-1]["locatie"]
        meta["tube_nr"] = int(float(metalist[-1]["filternummer"]))
        meta["name"] = "-".join([meta["monitoring_well"], metalist[-1]["filternummer"]])
        meta["x"] = float(metalist[-1]["x-coordinaat"])
        meta["y"] = float(metalist[-1]["y-coordinaat"])
        meetpunt = metalist[-1]["meetpunt nap"]
        if meetpunt == "":
            meta["tube_top"] = np.nan
        else:
            meta["tube_top"] = float(meetpunt) / 100.0
        maaiveld = metalist[-1]["maaiveld nap"]
        if maaiveld == "":
            meta["ground_level"] = np.nan
        else:
            meta["ground_level"] = float(maaiveld) / 100
        bovenkant_filter = metalist[-1]["bovenkant filter"]
        if bovenkant_filter == "":
            meta["screen_top"] = np.nan
        else:
            meta["screen_top"] = float(bovenkant_filter) / 100
        onderkant_filter = metalist[-1]["onderkant filter"]
        if onderkant_filter == "":
            meta["screen_bottom"] = np.nan
        else:
            meta["screen_bottom"] = float(onderkant_filter) / 100
        meta["metadata_available"] = True
    else:
        # no metadata
        meta["monitoring_well"] = ""
        meta["tube_nr"] = np.nan
        meta["name"] = "unknown"
        meta["x"] = np.nan
        meta["y"] = np.nan
        meta["tube_top"] = np.nan
        meta["ground_level"] = np.nan
        meta["screen_top"] = np.nan
        meta["screen_bottom"] = np.nan
        meta["metadata_available"] = False

    return line, meta


def _read_artdino_groundwater_measurements(f, line):
    line = line.strip()
    titel = line.split(",")
    while "" in titel:
        titel.remove("")

    if line != "":
        # Validate if titles are valid names
        validator = np.lib._iotools.NameValidator()
        titel = [s.lower() for s in validator(titel)]
        usecols = range(0, len(titel))

        try:
            measurements = pd.read_csv(
                f,
                header=None,
                names=titel,
                parse_dates=["peil_datum_tijd"],
                index_col="peil_datum_tijd",
                dayfirst=True,
                usecols=usecols,
            )
        except pd.errors.ParserError:
            # for now the workflow is to remove the files that cannot be read
            # manually.
            measurements = None
    else:
        measurements = None

    return line, measurements


def read_artdino_groundwater_csv(fname, to_mnap=True, read_series=True):
    """Read dino groundwater quantity data from a CSV file as exported by
    ArtDiver.

    Parameters
    ----------
    fname : str
        path to csv file
    to_mnap : boolean, optional
        if True a column with 'stand_m_tov_nap' is added to the dataframe
    read_series : boolean, optional
        if False only metadata is read, default is True

    Returns
    -------
    measurements : pd.DataFrame
    meta : dict
        dictionary with metadata
    """
    logger.info(f"reading -> {os.path.split(fname)[-1]}")

    with open(fname, "r") as f:
        # read header
        line, header = _read_dino_groundwater_header(f)
        line = _read_empty(f, line)
        if not header:
            logger.warning(f"could not read header -> {fname}")

        # read metadata
        line, meta = _read_artdino_groundwater_metadata(f, line)
        line = _read_empty(f, line)

        if not meta["metadata_available"]:
            logger.warning(f"could not read metadata -> {fname}")

        meta["filename"] = fname
        meta["source"] = "dino"

        # read measurements
        if read_series:
            line, measurements = _read_artdino_groundwater_measurements(f, line)
            if measurements is None:
                logger.warning(f"could not read measurements -> {fname}")
            elif measurements[~measurements["stand_cm_nap"].isna()].empty:
                logger.warning(f"no NAP measurements available -> {fname}")
            if to_mnap and measurements is not None:
                measurements["stand_m_tov_nap"] = measurements["stand_cm_nap"] / 100.0
                meta["unit"] = "m NAP"
            elif not to_mnap:
                meta["unit"] = "cm NAP"
        else:
            measurements = None

    return measurements, meta


def read_artdino_dir(
    dirname,
    ObsClass=None,
    subdir="csv",
    suffix=".csv",
    unpackdir=None,
    force_unpack=False,
    preserve_datetime=False,
    keep_all_obs=True,
    **kwargs,
):
    """Read Dino directory with point observations.

    TODO:
    - Evt. nog verbeteren door meteen Dataframe te vullen op het moment dat een
    observatie wordt ingelezen. Nu wordt eerst alles ingelezen in een lijst en
    daar een dataframe van gemaakt.
    - aparte unzip functie maken en toch de juiste tijdelijke directory krijgen.

    Parameters
    ----------
    dirname : str
        directory name, can be a .zip file or the parent directory of subdir
    ObsClass : type
        class of the observations, e.g. GroundwaterObs or WaterlvlObs
    subdir : str
        subdirectory of dirname with data files
    suffix : str
        suffix of files in subdir that will be read
    unpackdir : str
        destination directory of the unzipped file
    force_unpack : boolean, optional
        force unpack if dst already exists
    preserve_datetime : boolean, optional
        use date of the zipfile for the destination file
    keep_all_obs : boolean, optional
        add all observation points to the collection, even without data or
        metadata
    **kwargs: dict, optional
        Extra arguments are passed to ObsClass.from_dino_file()

    Returns
    -------
    obs_df : pd.DataFrame
        collection of multiple point observations
    """

    # unzip dir
    if dirname.endswith(".zip"):
        zipf = dirname
        if unpackdir is None:
            dirname = tempfile.TemporaryDirectory().name
        else:
            dirname = unpackdir
        unzip_file(
            zipf, dirname, force=force_unpack, preserve_datetime=preserve_datetime
        )

    # read filenames
    files = os.listdir(os.path.join(dirname, subdir))
    if suffix:
        files = [file for file in files if file.endswith(suffix)]

    if not files:
        raise FileNotFoundError(
            "no files were found in {} that end with {}".format(
                os.path.join(dirname, subdir), suffix
            )
        )

    # read individual files
    obs_list = []
    for _, file in enumerate(files):
        fname = os.path.join(dirname, subdir, file)
        obs = ObsClass.from_artdino_file(fname=fname, **kwargs)
        if obs.metadata_available and (not obs.empty):
            obs_list.append(obs)
        elif keep_all_obs:
            obs_list.append(obs)
        else:
            logging.info(f"not added to collection -> {fname}")

    return obs_list


def _read_dino_waterlvl_metadata(f, line):
    """read dino waterlevel metadata

    Parameters
    ----------
    f : text wrapper
    line : str
        line with meta dictionary keys
    meta_dic : dict (optional)
        dictionary with metadata

    Returns
    -------
    meta : dict
        dictionary with metadata

    """
    meta_keys = line.strip().split(",")
    meta_values = f.readline().strip().split(",")
    meta = {}
    for key, value in zip(meta_keys, meta_values):
        key = key.strip()
        if key in ["X-coordinaat", "Y-coordinaat"]:
            if key == "X-coordinaat":
                meta["x"] = float(value)
            elif key == "Y-coordinaat":
                meta["y"] = float(value)
        elif key == "Locatie":
            meta["monitoring_well"] = value
            meta["name"] = value

    return meta


def _read_dino_waterlvl_measurements(f, line):
    """

    Parameters
    ----------
    f : text wrapper
    line: str
        header of dataframe

    Returns
    -------
    measurements : pd.DataFrame
    """

    titel = line.strip().split(",")
    while "" in titel:
        titel.remove("")

    validator = np.lib._iotools.NameValidator()
    titel = [i.lower() for i in validator(titel)]
    usecols = range(0, len(titel))

    measurements = pd.read_csv(
        f,
        header=None,
        names=titel,
        parse_dates=["peildatum"],
        index_col="peildatum",
        dayfirst=True,
        usecols=usecols,
    )

    # measurements['Stand (m t.o.v. NAP)'] = measurements['Stand (cm t.o.v. NAP)'] /100.

    # measurements.set_index('Peildatum', inplace=True)

    return measurements


def read_dino_waterlvl_csv(fname, to_mnap=True, read_series=True):
    """Read dino waterlevel data from a dinoloket csv file.

    Parameters
    ----------
    fname : str
    to_mnap : boolean, optional
        if True a column with 'stand_m_tov_nap' is added to the dataframe
    read_series : boolean, optional
        if False only metadata is read, default is True
    """
    logging.info(f"reading -> {os.path.split(fname)[-1]}")

    p_meta = re.compile(
        "Locatie,Externe aanduiding,X-coordinaat,Y-coordinaat, Startdatum, Einddatum"
    )
    p_data = re.compile(r"Locatie,Peildatum,Stand \(cm t.o.v. NAP\),Bijzonderheid")

    with open(fname, "r") as f:
        line = f.readline()
        while line != "":
            line = f.readline()
            if p_meta.match(line):
                meta = _read_dino_waterlvl_metadata(f, line)
                if meta:
                    meta["metadata_available"] = True
                else:
                    meta["metadata_available"] = False
                meta["filename"] = fname
                meta["source"] = "dino"
            elif p_data.match(line):
                if read_series:
                    measurements = _read_dino_waterlvl_measurements(f, line)
                    if to_mnap and measurements is not None:
                        measurements["stand_m_tov_nap"] = (
                            measurements["stand_cm_tov_nap"] / 100.0
                        )
                        meta["unit"] = "m NAP"
                    elif not to_mnap:
                        meta["unit"] = "cm NAP"
                else:
                    measurements = None

                return measurements, meta


def read_dino_dir(
    dirname,
    ObsClass=None,
    subdir="Boormonsterprofiel_Geologisch booronderzoek",
    suffix=".txt",
    unpackdir=None,
    force_unpack=False,
    preserve_datetime=False,
    keep_all_obs=True,
    **kwargs,
):
    """Read Dino directory with point observations.

    TODO:
    - Evt. nog verbeteren door meteen Dataframe te vullen op het moment dat een
    observatie wordt ingelezen. Nu wordt eerst alles ingelezen in een lijst en
    daar een dataframe van gemaakt.
    - aparte unzip functie maken en toch de juiste tijdelijke directory krijgen.

    Parameters
    ----------
    dirname : str
        directory name, can be a .zip file or the parent directory of subdir
    ObsClass : type
        class of the observations, e.g. GroundwaterObs or WaterlvlObs
    subdir : str
        subdirectory of dirname with data files
    suffix : str
        suffix of files in subdir that will be read
    unpackdir : str
        destination directory of the unzipped file
    force_unpack : boolean, optional
        force unpack if dst already exists
    preserve_datetime : boolean, optional
        use date of the zipfile for the destination file
    keep_all_obs : boolean, optional
        add all observation points to the collection, even without data or
        metadata
    **kwargs: dict, optional
        Extra arguments are passed to ObsClass.from_dino_file()

    Returns
    -------
    obs_df : pd.DataFrame
        collection of multiple point observations
    """

    # unzip dir
    if dirname.endswith(".zip"):
        zipf = dirname
        if unpackdir is None:
            dirname = tempfile.TemporaryDirectory().name
        else:
            dirname = unpackdir
        unzip_file(
            zipf, dirname, force=force_unpack, preserve_datetime=preserve_datetime
        )

    # read filenames
    files = os.listdir(os.path.join(dirname, subdir))
    if suffix:
        files = [file for file in files if file.endswith(suffix)]

    if not files:
        raise FileNotFoundError(
            "no files were found in {} that end with {}".format(
                os.path.join(dirname, subdir), suffix
            )
        )

    # read individual files
    obs_list = []
    for _, file in enumerate(files):
        fname = os.path.join(dirname, subdir, file)
        obs = ObsClass.from_dino(fname=fname, **kwargs)
        if obs.metadata_available and (not obs.empty):
            obs_list.append(obs)
        elif keep_all_obs:
            obs_list.append(obs)
        else:
            logging.info(f"not added to collection -> {fname}")

    return obs_list



import os
import zipfile

import numpy as np
import pandas as pd
from tqdm import tqdm


def read_waterinfo_file(
    path_to_file,
    index_cols=None,
    return_metadata=False,
    value_col=None,
    location_col=None,
    xcol=None,
    ycol=None,
    transform_coords=True,
):
    """Read waterinfo file (CSV or zip)

    Parameters
    ----------
    path_to_file : str
        path to file

    Returns
    -------
    df : pandas.DataFrame
        DataFrame containing file content
    metadata : dict, optional
        dict containing metadata, returned if return_metadata is
        True, default is False
    """
    from pyproj import Transformer

    name = os.path.splitext(os.path.basename(path_to_file))[0]

    if path_to_file.endswith(".csv"):
        f = path_to_file
    elif path_to_file.endswith(".zip"):
        zf = zipfile.ZipFile(path_to_file)
        f = zf.open("{}.csv".format(name))
    else:
        raise NotImplementedError(
            "File type '{}' not supported!".format(os.path.splitext(path_to_file)[-1])
        )

    if index_cols is None:
        index_cols = ["WAARNEMINGDATUM", "WAARNEMINGTIJD"]

    if value_col is None:
        value_col = "NUMERIEKEWAARDE"

    if location_col is None:
        location_col = "MEETPUNT_IDENTIFICATIE"

    if xcol is None:
        xcol = "X"

    if ycol is None:
        ycol = "Y"

    # read data
    df = pd.read_csv(
        f,
        sep=";",
        decimal=",",
        encoding="ISO-8859-1",
        parse_dates=[index_cols],
        dayfirst=True,
        infer_datetime_format=True,
        index_col="_".join(index_cols),
    )

    # do some conversions
    df.loc[df[value_col] == 999999999, value_col] = np.NaN
    df[value_col] = df[value_col] / 100.0

    # parse metadata into dict
    if return_metadata:
        if len(df[location_col].unique()) > 1:
            raise ValueError(
                "File contains data for more than one location!"
                " Use ObsCollection.from_waterinfo()!"
            )

        metadata = {}

        if transform_coords:
            transformer = Transformer.from_crs("epsg:25831", "epsg:28992")
            x, y = transformer.transform(df[xcol].iloc[-1], df[ycol].iloc[-1])
        else:
            x = df[xcol].iloc[-1] / 100.0
            y = df[ycol].iloc[-1] / 100.0
        metadata["name"] = df[location_col].iloc[-1]
        metadata["x"] = x
        metadata["y"] = y
        metadata["filename"] = f
        metadata["source"] = "waterinfo"

        return df, metadata
    else:
        return df


def read_waterinfo_obs(file_or_dir, ObsClass, progressbar=False, **kwargs):
    """Read waterinfo file or directory and extract locations and observations.

    Parameters
    ----------
    file_or_dir : str
        path to file or directory
    ObsClass: Obs type
        type of Obs to store data in
    progressbar : bool, optional
        show progressbar if True, default is False

    Returns
    -------
    obs_collection : list
        list of Obs objects
    """
    from pyproj import Transformer

    # Waterinfo file
    if os.path.isfile(file_or_dir):
        files = [file_or_dir]
    # directory with waterinfo files (zips or csvs)
    elif os.path.isdir(file_or_dir):
        files = [os.path.join(file_or_dir, f) for f in sorted(os.listdir(file_or_dir))]
    else:
        raise NotImplementedError("Provide path to file or directory!")

    location_col = kwargs.pop("location_col", "MEETPUNT_IDENTIFICATIE")

    # loop over files
    metadata = {}
    obs_collection = []

    transformer = Transformer.from_crs("epsg:25831", "epsg:28992")

    for filenm in tqdm(files) if progressbar else files:
        # read file or zip
        df = read_waterinfo_file(filenm, location_col=location_col, **kwargs)

        # get location and convert to m RD
        for stn in df[location_col].unique():
            mask = df[location_col] == stn
            x, y = transformer.transform(df.loc[mask, "X"][-1], df.loc[mask, "Y"][-1])
            metadata = {
                "name": stn,
                "x": x,
                "y": y,
                "filename": filenm,
                "source": "waterinfo",
            }

            # add to list
            o = ObsClass(df.loc[mask, :], meta=metadata, **metadata)
            obs_collection.append(o)

    return obs_collection



import logging
import os

import numpy as np
import pandas as pd

from ..util import get_files

logger = logging.getLogger(__name__)


def _read_wiski_header(f, header_sep=":", header_identifier="#", end_header_str=None):
    line = f.readline()
    header = dict()
    while header_identifier in line:
        prop, val = line.split(header_sep)
        prop = prop.strip().replace(header_identifier, "")
        val = val.strip()
        try:
            val = float(val)
        except ValueError:
            pass
        header[prop] = val
        line = f.readline()
        if end_header_str is not None:
            if end_header_str in line:
                break

    return line, header


def read_wiski_file(
    fname,
    sep=";",
    header_sep=None,
    header_identifier="#",
    read_series=True,
    infer_datetime_format=True,
    translate_dic=None,
    tz_localize=True,
    unit="",
    **kwargs,
):
    """
    Read data from a WISKI file.

    Parameters:
    -----------
    fname : str
        The name of the file to be read.
    sep : str, optional (default=";")
        The delimiter used to separate fields in the file.
    header_sep : str, optional (default=None)
        The delimiter used to separate fields in the header. If None, the
        function will try to automatically detect the separator.
    header_identifier : str, optional (default="#")
        The character used to identify header lines.
    read_series : bool, optional (default=True)
        Whether to read the time series data from the file.
    infer_datetime_format : bool, optional (default=True)
        Whether to infer the datetime format of the timestamp column.
    translate_dic : dict, optional (default=None)
        A dictionary mapping header field names to the desired output names.
    tz_localize : bool, optional (default=True)
        Whether to localize the datetime index to the machine's timezone.
    unit : str, optional (default="")
        The unit of measurement of the data.
    **kwargs : keyword arguments
        Additional arguments to pass to the pandas `read_csv` function.

    Returns:
    --------
    data : pandas.DataFrame or None
        A dataframe containing the time series data from the file. Returns None
        if `read_series` is False.
    metadata : dict
        A dictionary containing metadata about the data in the file.
    """
    logger.info("reading -> {}".format(os.path.split(fname)[-1]))

    if translate_dic is None:
        translate_dic = {}

    # manually break header parse at certain point
    if "end_header_str" in kwargs.keys():
        end_header_str = kwargs.pop("end_header_str")
    else:
        end_header_str = None

    # read header
    with open(fname, "r") as f:
        if header_sep is None:
            line, header = _read_wiski_header(
                f, end_header_str=end_header_str, header_identifier=header_identifier
            )
        else:
            line, header = _read_wiski_header(
                f,
                header_sep=header_sep,
                header_identifier=header_identifier,
                end_header_str=end_header_str,
            )

        # specify names
        for key, item in translate_dic.items():
            header[key] = header[item]

        # get column names of data
        # columns = split(r'\s{2,}', line)
        if sep == r"\s+":
            columns = line.split("\t")
        else:
            columns = line.split(sep)
        columns = [icol.strip("\n") for icol in columns]
        columns = [icol.replace("[", "_[") for icol in columns]

        validator = np.lib._iotools.NameValidator()
        columns = validator(columns)

        if read_series:
            # read data
            data = pd.read_csv(
                f,
                sep=sep,
                header=None,
                names=columns,
                infer_datetime_format=infer_datetime_format,
                **kwargs,
            )

            if tz_localize:
                data.index = data.index.tz_localize(None)

            # convert Value to float
            col = [icol for icol in data.columns if icol.lower().startswith("value")][0]

            data[col] = pd.to_numeric(data[col], errors="coerce")
        else:
            data = None

        # translate some header keys
        metadata = {"source": "wiski", "unit": unit}
        for key, val in header.items():
            if key == "Station Site":
                metadata["monitoring_well"] = val
            elif key == "x":
                metadata["x"] = val
            elif key == "y":
                metadata["y"] = val
            elif key == "name":
                metadata["name"] = val
            # this adds the other header keys to metadata
            # but this will not work if keys contain spaces etc.
            # because it tries adding them as attributes.
            # else:
            #     metadata[key] = val

    return data, metadata


def read_wiski_dir(
    dirname,
    ObsClass=None,
    suffix=".csv",
    unpackdir=None,
    force_unpack=False,
    preserve_datetime=False,
    keep_all_obs=True,
    **kwargs,
):
    """
    Reads WISKI CSV files from a directory and returns a list of observation
    objects.

    Parameters
    ----------
    dirname : str
        The path of the directory containing the WISKI CSV files.
    ObsClass : object, optional
        The observation class to use for creating observation objects. Default
        is None.
    suffix : str, optional
        The file extension of the WISKI CSV files. Default is ".csv".
    unpackdir : str, optional
        The directory to which the files should be unpacked. Default is None.
    force_unpack : bool, optional
        If True, forces the files to be unpacked even if they are already in the
        target directory. Default is False.
    preserve_datetime : bool, optional
        If True, preserves the original modification times of the files when
        unpacking them. Default is False.
    keep_all_obs : bool, optional
        If True, keeps all observation objects even if they have no metadata
        available. Default is True.
    **kwargs
        Additional keyword arguments to pass to the `from_wiski` method of the
        `ObsClass` object.

    Returns
    -------
    list
        A list of observation objects created from the WISKI CSV files in the
        directory.

    Raises
    ------
    FileNotFoundError
        If no WISKI CSV files are found in the directory.

    """
    # get files
    dirname, unzip_fnames = get_files(
        dirname,
        ext=suffix,
        force_unpack=force_unpack,
        preserve_datetime=preserve_datetime,
    )

    if not unzip_fnames:
        raise FileNotFoundError(
            "no files were found in '{}' that end with '{}'".format(
                os.path.join(dirname), suffix
            )
        )

    # gather all obs in list
    obs_list = []
    for i, csv in enumerate(unzip_fnames):
        logger.info("reading {0}/{1} -> {2}".format(i + 1, len(unzip_fnames), csv))
        obs = ObsClass.from_wiski(os.path.join(dirname, csv), **kwargs)

        if obs.metadata_available:
            obs_list.append(obs)
        elif keep_all_obs:
            obs_list.append(obs)
        else:
            logger.info(f"not added to collection -> {csv}")

    return obs_list



# -*- coding: utf-8 -*-
"""Created on Wed Sep 12 12:15:42 2018.

@author: Artesia
"""

import logging
import os
import sys
import tempfile
import time
import zipfile
from typing import Dict, List, Optional

from colorama import Back, Fore, Style
from numpy import unique
from pandas import DataFrame, DatetimeIndex
from scipy.interpolate import RBFInterpolator

logger = logging.getLogger(__name__)


def _obslist_to_frame(obs_list):
    """convert a list of observations to a pandas DataFrame.

    Parameters
    ----------
    obs_list : list of hydropandas.*Obs
        list containing *Obs objects that will be stored in DataFrame.

    Returns
    -------
    obs_df : pandas.DataFrame
        DataFrame containing all data
    """
    if len(obs_list) > 0:
        obs_df = DataFrame(
            [o.to_collection_dict() for o in obs_list],
            columns=obs_list[0].to_collection_dict().keys(),
        )
        obs_df.set_index("name", inplace=True)
        if obs_df.index.duplicated().any():
            logger.warning("multiple observations with the same name")
    else:
        obs_df = DataFrame()

    return obs_df


def unzip_file(src, dst, force=False, preserve_datetime=False):
    """Unzip file.

    Parameters
    ----------
    src : str
        source zip file
    dst : str
        destination directory
    force : boolean, optional
        force unpack if dst already exists
    preserve_datetime : boolean, optional
        use date of the zipfile for the destination file

    Returns
    -------
    int
        1 of True
    """
    if os.path.exists(dst):
        if not force:
            print(
                "File not unzipped. Destination already exists. Use"
                "'force=True' to unzip."
            )
            return
    if preserve_datetime:
        zipf = zipfile.ZipFile(src, "r")
        for f in zipf.infolist():
            zipf.extract(f, path=dst)
            date_time = time.mktime(f.date_time + (0, 0, -1))
            os.utime(os.path.join(dst, f.filename), (date_time, date_time))
        zipf.close()
    else:
        zipf = zipfile.ZipFile(src, "r")
        zipf.extractall(dst)
        zipf.close()
    return 1


def get_files(
    file_or_dir, ext, unpackdir=None, force_unpack=False, preserve_datetime=False
):
    """internal method to get list of files with specific extension from
    dirname.

    Parameters
    ----------
    file_or_dir : str
        file or path to data
    ext : str
        extension of filenames to store in list
    force_unpack : bool, optional
        force unzip, by default False
    preserve_datetime : bool, optional
        preserve datetime of unzipped files, by default False
        (useful for checking whether data has changed)
    """
    # check if unpackdir is same as file_or_dir, if same, this can cause
    # problems when the unpackdir still contains zips that will be unpacked
    # again.
    if unpackdir is not None:
        if os.path.normcase(unpackdir) == os.path.normcase(file_or_dir):
            raise ValueError("Please specify a different folder to unpack" " files!")
    # identify whether file_or_dir started as zip
    if file_or_dir.endswith(".zip"):
        iszip = True
    else:
        iszip = False

    # unzip dir
    if file_or_dir.endswith(".zip"):
        zipf = file_or_dir
        if unpackdir is None:
            file_or_dir = tempfile.TemporaryDirectory().name
        else:
            file_or_dir = unpackdir
        unzip_file(
            zipf, file_or_dir, force=force_unpack, preserve_datetime=preserve_datetime
        )

    # file_or_dir is directory
    if os.path.isdir(file_or_dir):
        # check for zips in dir
        zip_fnames = [i for i in os.listdir(file_or_dir) if i.endswith(".zip")]
        if len(zip_fnames) > 0:
            # unzip zips
            if unpackdir is None:
                dirname = tempfile.TemporaryDirectory().name
            else:
                dirname = unpackdir
            for zipf in zip_fnames:
                unzip_file(
                    os.path.join(file_or_dir, zipf),
                    dirname,
                    force=True,
                    preserve_datetime=preserve_datetime,
                )
                # remove intermediate zipfiles if initial file_or_dir was zip
                if iszip:
                    os.remove(os.path.join(file_or_dir, zipf))
        else:
            dirname = file_or_dir
        # get all files with extension ext
        unzip_fnames = [i for i in os.listdir(dirname) if i.endswith(ext)]
    elif os.path.isfile(file_or_dir):
        # file_or_dir is actually an xml
        unzip_fnames = [os.path.basename(file_or_dir)]  # get file name
        dirname = os.path.dirname(file_or_dir)  # get directory path
    else:
        raise NotImplementedError("Cannot parse 'file_or_dir': " f"{file_or_dir}!")

    return dirname, unzip_fnames


def df2gdf(df, xcol="x", ycol="y"):
    """Make a GeoDataFrame from a DataFrame, assuming the geometry are
    points."""
    from geopandas import GeoDataFrame
    from shapely.geometry import Point

    gdf = GeoDataFrame(
        df.copy(), geometry=[Point((s[xcol], s[ycol])) for i, s in df.iterrows()]
    )
    return gdf


def show_versions():
    """Method to print the version of dependencies."""
    from sys import version as os_version

    from matplotlib import __version__ as mpl_version
    from numpy import __version__ as np_version
    from pandas import __version__ as pd_version
    from scipy import __version__ as sc_version

    msg = (
        f"Python version    : {os_version}\n"
        f"Numpy version     : {np_version}\n"
        f"Scipy version     : {sc_version}\n"
        f"Pandas version    : {pd_version}\n"
        f"Matplotlib version: {mpl_version}"
    )

    return print(msg)


class ColoredFormatter(logging.Formatter):
    """Colored log formatter.

    Taken from
    https://gist.github.com/joshbode/58fac7ababc700f51e2a9ecdebe563ad
    """

    def __init__(
        self, *args, colors: Optional[Dict[str, str]] = None, **kwargs
    ) -> None:
        """Initialize the formatter with specified format strings."""

        super().__init__(*args, **kwargs)

        self.colors = colors if colors else {}

    def format(self, record) -> str:
        """Format the specified record as text."""

        record.color = self.colors.get(record.levelname, "")
        record.reset = Style.RESET_ALL

        return super().format(record)


def get_color_logger(level="INFO"):
    formatter = ColoredFormatter(
        "{color}{levelname}:{name}:{message}{reset}",
        style="{",
        datefmt="%Y-%m-%d %H:%M:%S",
        colors={
            "DEBUG": Fore.CYAN,
            "INFO": Fore.GREEN,
            "WARNING": Fore.YELLOW,
            "ERROR": Fore.RED,
            "CRITICAL": Fore.RED + Back.WHITE + Style.BRIGHT,
        },
    )

    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)

    logger = logging.getLogger()
    logger.handlers[:] = []
    logger.addHandler(handler)
    logger.setLevel(getattr(logging, level))

    logging.captureWarnings(True)
    return logger


def oc_to_df(oc, col: Optional[str] = None) -> DataFrame:
    obs_times = DatetimeIndex(unique([obs.index for obs in oc.obs]))
    df = DataFrame(index=obs_times, columns=oc.index, dtype="float").sort_index()
    for obs in oc.obs:
        if not obs.empty:
            if col is None:
                vals = obs.loc[:, obs._get_first_numeric_col_name()]
            else:
                vals = obs.loc[:, col]
            df.loc[obs.index, obs.name] = vals.values

    return df


def interpolate(
    xy: List[List[float]],
    obsdf: DataFrame,
    obsloc: DataFrame,
    kernel: str = "thin_plate_spline",
    kernel2: str = "linear",
    epsilon: Optional[int] = None,
) -> DataFrame:
    """Interpolation method using the Scipy radial basis function (RBF)


    Parameters
    ----------
    xy : List[List[float]]
        xy coordinates of locations of interest e.g. [[10,25], [5,25]]
    obsdf : DataFrame
        Dataframe containing the observation locations as columns and
        the observations at a measurement time in each row.
    obsloc : DataFrame
        Dataframe containing the observation locations coordinates
        with observation locations as index and columns ["x", "y"]
    kernel : str, optional
        Type of radial basis funtion, by default thin_plate_spline.
        Other options are linear, gaussian, inverse_quadratic,
        multiquadric, inverse_multiquadric, cubic or quintic.
    kernel2 : str, optional
        Kernel in case there are not enough observations (3 or 6) for
        time step, by default linear. Other options are gaussian,
        inverse_quadratic, multiquadric, or inverse_multiquadric.
    epsilon : Optional[int], optional
        Shape parameter that scales the input to the RBF. If kernel is
        linear, thin_plate_spline, cubic, or quintic, this defaults to 1.
        Otherwise this must be specified.

    Returns
    -------
    DataFrame
        DataFrame with locations of interest as columns and interpolated values
        at a measurement time in each row.
    """

    if (kernel == "thin_plate_spline") or (kernel == "cubic"):
        min_val = 3
    elif kernel == "quintic":
        min_val = 6
    else:
        min_val = len(obsdf.index)

    fill_df = (
        DataFrame(
            index=obsdf.index,
            columns=[f"{int(_xy[0])}_{int(_xy[1])}" for _xy in xy],
        )
        .sort_index()
        .astype(float)
    )

    for idx in obsdf.index:
        # get all stations with values for this date
        val = obsdf.loc[idx].dropna()
        # get stations for this date
        coor = obsloc.loc[val.index]

        if len(val) >= min_val:
            kernel = kernel
        else:
            kernel = kernel2

        # create an scipy interpolator
        rbf = RBFInterpolator(
            coor.to_numpy(), val.to_numpy(), epsilon=epsilon, kernel=kernel
        )

        val_rbf = rbf(xy)
        fill_df.loc[idx] = val_rbf

    return fill_df



import logging

import pandas as pd

import hydropandas as hpd
from hydropandas.io import knmi

logging.basicConfig(level=logging.DEBUG)


def test_read_knmi_file1():
    # De Bilt neerslagstation
    fname = "./tests/data/2023-KNMI-test/neerslaggeg_ESBEEK_831.txt"
    knmi.read_knmi_daily_rainfall_file(fname)

    return


def test_read_knmi_file2():
    # De Bilt neerslagstation
    fname = "./tests/data/2023-KNMI-test/neerslaggeg_VILSTEREN_342.txt"
    knmi.read_knmi_daily_rainfall_file(fname)

    return


def test_read_knmi_file3():
    # De Bilt neerslagstation
    fname = "./tests/data/2023-KNMI-test/precipitation_st_anthonis.txt"
    knmi.read_knmi_daily_rainfall_file(fname)

    return


def test_get_knmi_precip_neerslagstation():
    # De Bilt neerslagstation
    knmi.get_knmi_timeseries_stn(
        "550", "RD", start="2010-1-1", end="2010-1-10", settings=None
    )
    return


def test_read_knmi_meteostation_file():
    fname = "./tests/data/2023-KNMI-test/etmgeg_260.txt"
    knmi.read_knmi_daily_meteo_file(fname, "EV24")

    return


def test_get_knmi_precip_meteostation_fill_missing():
    # De Bilt meteostation
    knmi.get_knmi_timeseries_stn(
        260, "RH", start="2010-1-1", end="2010-1-10", settings=None
    )
    return


def test_get_knmi_precip_meteostation_hourly():
    # De Bilt meteostation uurlijks
    knmi.get_knmi_timeseries_stn(
        260,
        "RH",
        start="2010-1-1",
        end="2010-1-10",
        settings={"interval": "hourly", "fill_missing_obs": False},
    )
    return


def test_get_pressure_hourly():
    # De Bilt meteostation uurlijks
    knmi.get_knmi_timeseries_stn(
        310,
        "P",
        start="2010-1-1",
        end="2010-1-10",
        settings={"interval": "hourly", "fill_missing_obs": False},
    )
    return


def test_get_knmi_precip_neerslagstation_no_api():
    # De Bilt neerslagstation
    knmi.get_knmi_timeseries_stn(
        "550", "RD", start="2010-1-1", end="2010-1-10", settings={"use_api": False}
    )
    return


def test_get_knmi_precip_meteostation_no_api():
    # De Bilt neerslagstation
    knmi.get_knmi_timeseries_stn(
        260, "RH", start="2010-1-1", end="2010-1-10", settings={"use_api": False}
    )
    return


def test_get_knmi_precip_neerslagstation_fill_missing():
    settings = knmi._get_default_settings(None)
    knmi.download_knmi_data(
        "550", meteo_var="RD", start="1952", end=None, settings=settings
    )
    return


def test_download_rd_550_no_api():
    settings = knmi._get_default_settings(None)
    settings["use_api"] = False
    knmi.download_knmi_data(
        "550",
        stn_name="DE-BILT",
        meteo_var="RD",
        start="1952",
        end=None,
        settings=settings,
    )
    return


def test_download_rd_12():
    settings = knmi._get_default_settings(None)
    knmi.download_knmi_data(
        "12", meteo_var="RD", start="2010", end=None, settings=settings
    )
    return


def test_download_without_data():
    settings = knmi._get_default_settings(None)
    try:
        knmi.download_knmi_data(
            "324", meteo_var="RD", start="2018", end="2020", settings=settings
        )
    except ValueError:
        pass

    return


def test_download_without_data_no_error():
    settings = knmi._get_default_settings(settings={"raise_exceptions": False})

    knmi_df, variables, stations = knmi.download_knmi_data(
        "324", meteo_var="RD", start="2018", end="2020", settings=settings
    )

    assert (knmi_df.empty, variables == {}, stations.empty) == (True, True, True)

    return


def test_download_ev24_210():
    settings = knmi._get_default_settings(None)

    knmi.download_knmi_data(
        210, meteo_var="EV24", start="1952", end=None, settings=settings
    )
    return


def test_download_ev24_210_no_api():
    settings = knmi._get_default_settings({"use_api": False})
    knmi.download_knmi_data(
        210, meteo_var="EV24", start="1952", end=None, settings=settings
    )
    return


def test_get_knmi_daily_meteo_ev24_empty():
    start, end = knmi._start_end_to_datetime("1959", "1963")
    knmi.get_knmi_daily_meteo_api(knmi.URL_DAILY_METEO, 265, "EV24", start, end, False)

    return


def test_fill_missing_measurements_ev24_278():
    knmi.fill_missing_measurements(278, meteo_var="EV24", start="1959", end="1963")
    return


def test_fill_missing_measurements_rh_273():
    knmi.fill_missing_measurements(273, meteo_var="RH", start="1995", end="2000")
    return


def test_fill_missing_measurements_rh_278():
    settings = knmi._get_default_settings({"exclude_neerslag_stn": True})
    knmi.fill_missing_measurements(
        278, meteo_var="RH", start="1990", end=None, settings=settings
    )
    return


def test_obslist_from_grid():
    xy = [[x, y] for x in [104150.0, 104550.0] for y in [510150.0, 510550.0]]

    knmi.get_knmi_obslist(
        xy=xy,
        meteo_vars=["RH", "EV24"],
        starts=["2010", "2010"],
        ObsClasses=[hpd.PrecipitationObs, hpd.EvaporationObs],
    )

    return


def test_obslist_from_locations():
    locations = pd.DataFrame(data={"x": [100000], "y": [350000]})
    knmi.get_knmi_obslist(
        locations, meteo_vars=["EV24"], ObsClasses=[hpd.EvaporationObs]
    )

    return


def test_obslist_from_stns():
    stns = [344, 260]  # Rotterdam en de Bilt
    knmi.get_knmi_obslist(
        stns=stns,
        meteo_vars=["RH", "EV24"],
        starts=["2010", "2010"],
        ends=["2015", "2015"],
        ObsClasses=[hpd.PrecipitationObs, hpd.EvaporationObs],
    )

    return


def test_obslist_from_stns_single_startdate():
    stns = [344, 260]  # Rotterdam en de Bilt
    knmi.get_knmi_obslist(
        stns=stns,
        meteo_vars=["RH", "EV24"],
        starts="2010",
        ends="2015",
        ObsClasses=[hpd.PrecipitationObs, hpd.EvaporationObs],
    )

    return



# -*- coding: utf-8 -*-
"""Created on Mon Jun 24 11:43:27 2019.

@author: oebbe
"""

import hydropandas as hpd
from hydropandas.io import wiski


def test_read_wiski_csv():
    # download single file

    wiski.read_wiski_file(
        "./tests/data/2019-WISKI-test/1016_PBF.csv",
        sep=r"\s+",
        header_sep=":",
        header_identifier=":",
        verbose=True,
        parse_dates={"datetime": [0, 1]},
        infer_datetime_format=True,
        index_col=["datetime"],
        translate_dic={"name": "Station Number", "x": "GlobalX", "y": "GlobalY"},
    )

    return


def test_read_wiski_csv2():
    # download single file

    wiski.read_wiski_file(
        "./tests/data/2019-WISKI-test/8137_PBF.csv",
        sep=r"\s+",
        header_sep=":",
        header_identifier=":",
        verbose=True,
        parse_dates={"datetime": [0, 1]},
        infer_datetime_format=False,
        dayfirst=True,
        index_col=["datetime"],
        translate_dic={"name": "Station Number", "x": "GlobalX", "y": "GlobalY"},
    )

    return


def test_read_wiski_zip():
    wiski.read_wiski_dir(
        "./tests/data/2019-WISKI-test/1016_PBF.zip",
        ObsClass=hpd.GroundwaterObs,
        sep=r"\s+",
        header_sep=":",
        header_identifier=":",
        parse_dates={"datetime": [0, 1]},
        index_col=["datetime"],
        translate_dic={"name": "Station Number", "x": "GlobalX", "y": "GlobalY"},
        verbose=True,
    )

    return


def test_rijnenijssel_wiski_format():
    hpd.GroundwaterObs.from_wiski(
        (
            "./tests/data/2019-WISKI-test/"
            "Zwiepse Horstweg Barchem_1024_FT1_WNS9040_MomentaanO.csv"
        ),
        header_sep=";",
        end_header_str="#Timestamp",
        parse_dates=[0],
        index_col=[0],
        infer_datetime_format=True,
        tz_localize=False,
    )
    return
